from collections.abc import Iterable
from numbers import Number
import numpy as np

import openmdao.api as om
from ..utils.misc import _unspecified, _none_or_unspecified


class ControlOptionsDictionary(om.OptionsDictionary):
    """
    An OptionsDictionary specific to controls.

    Parameters
    ----------
    read_only : bool
        If True, setting (via __setitem__ or update) is not permitted.
    """
    def __init__(self, read_only=False):
        super(ControlOptionsDictionary, self).__init__(read_only)

        self.declare(name='name', types=str,
                     desc='The name of ODE system parameter to be controlled.')

        self.declare(name='control_type', types=str, default='full',
                     desc='The type of control variable.  Options are `full` or `polynomial`.')

        self.declare(name='order', types=(int,), default=None, allow_none=True,
                     desc='A integer that provides the interpolation order when the control is '
                          'to assume a single polynomial basis across the entire phase, or None '
                          'to use the default control behavior.')

        self.declare(name='units', default=_unspecified,
                     allow_none=True, desc='The units in which the control variable is defined.')

        self.declare(name='desc', types=str, default='',
                     desc='The description of the control variable.')

        self.declare(name='opt', default=True, types=bool,
                     desc='If True, the control value will be a design variable '
                          'for the optimization problem.  If False, allow the '
                          'control to be connected externally.')

        self.declare(name='fix_initial', types=bool, default=False,
                     desc='If True, the initial value of this control is fixed and not a '
                          'design variable. This option is invalid if opt=False.')

        self.declare(name='fix_final', types=bool, default=False,
                     desc='If True, the final value of this control is fixed and not a '
                          'design variable. This option is invalid if opt=False.')

        self.declare(name='targets', allow_none=True, default=_unspecified,
                     desc='Targets in the ODE to which the state is connected')

        self.declare(name='rate_targets', allow_none=True, default=_unspecified,
                     desc='The targets in the ODE to which the control rate is connected')

        self.declare(name='rate2_targets', allow_none=True, default=_unspecified,
                     desc='The targets in the ODE to which the control 2nd derivative '
                          'is connected')

        self.declare(name='val', types=(Iterable, np.ndarray, Number), default=np.zeros(1),
                     desc='The default value of the control variable at the '
                          'control discretization nodes.')

        self.declare(name='shape', types=Iterable, allow_none=True, default=None,
                     desc='The shape of the control variable at each point in time.')

        self.declare(name='lower', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The lower bound of the control variable at the nodes. This '
                          'option is invalid if opt=False.')

        self.declare(name='upper', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The upper bound of the control variable at the nodes. This '
                          'option is invalid if opt=False.')

        self.declare(name='scaler', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The scaler of the control variable at the nodes. This '
                          'option is invalid if opt=False.')

        self.declare(name='adder', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The adder of the control variable at the nodes. This '
                          'option is invalid if opt=False.')

        self.declare(name='ref0', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The zero-reference value of the control variable at the nodes. This '
                          'option is invalid if opt=False.')

        self.declare(name='ref', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The unit-reference value of the control variable at the nodes. This '
                          'option is invalid if opt=False.')

        self.declare(name='continuity', types=(bool, dict), default=True,
                     desc='Enforce continuity of control values at segment boundaries. This '
                          'option is invalid if opt=False.')

        self.declare(name='continuity_scaler', types=(Number,), default=None, allow_none=True,
                     desc='Scaler for continuity at segment boundaries. This '
                          'option is invalid if opt=False.')

        self.declare(name='continuity_ref', types=(Number,), default=None, allow_none=True,
                     desc='Reference unit value for continuity at segment boundaries instead of scaler.'
                          'This option is invalid if opt=False.')

        self.declare(name='rate_continuity', types=(bool, dict), default=True,
                     desc='Enforce continuity of control first derivatives in dimensionless time '
                          'at segment boundaries. '
                          'This option is invalid if opt=False.')

        self.declare(name='rate_continuity_scaler', types=(Number,), default=None, allow_none=True,
                     desc='Scaler of the rate continuity constraint at segment boundaries. '
                          'This option is invalid if opt=False.')

        self.declare(name='rate_continuity_ref', types=(Number,), default=None, allow_none=True,
                     desc='Reference unit value for rate continuity at segment boundaries instead of scaler.'
                          'This option is invalid if opt=False.')

        self.declare(name='rate2_continuity', types=(bool, dict), default=False,
                     desc='Enforce continuity of control second derivatives at segment boundaries. '
                          'This option is invalid if opt=False.')

        self.declare(name='rate2_continuity_scaler', types=(Number,), default=None, allow_none=True,
                     desc='Scaler of the rate2 continuity constraint at segment boundaries. '
                          'This option is invalid if opt=False.')

        self.declare(name='rate2_continuity_ref', types=(Number,), default=None, allow_none=True,
                     desc='Reference unit value for rate2 continuity at segment boundaries instead of scaler.'
                          'This option is invalid if opt=False.')


def check_valid_shape(name, value):
    """
    Raise an exception if the value specified for a shape is invalid.

    Parameters
    ----------
    name : str
        Name of the option.
    value : object
        Shape to check, should be a Iterable, Number, list, or tuple.
    """
    if name == 'shape':
        if value not in _none_or_unspecified and not isinstance(value, (Iterable, Number, list, tuple)):
            raise ValueError(f"Option '{name}' with value {value} is not valid.")


class ParameterOptionsDictionary(om.OptionsDictionary):
    """
    An OptionsDictionary specific to parameters.

    Parameters
    ----------
    read_only : bool
        If True, setting (via __setitem__ or update) is not permitted.
    """
    def __init__(self, read_only=False):
        super(ParameterOptionsDictionary, self).__init__(read_only)

        self.declare(name='name', types=str,
                     desc='The name of ODE system parameter to be set via parameter.')

        self.declare(name='units', default=_unspecified,
                     allow_none=True, desc='The units in which the parameter is defined.')

        self.declare(name='desc', types=str, default='',
                     desc='The description of the parameter.')

        self.declare(name='opt', default=True, types=bool,
                     desc='If True, the control value will be a design variable '
                          'for the optimization problem.  If False, allow the '
                          'control to be connected externally.')

        self.declare(name='dynamic', values=[True, False, _unspecified], default=_unspecified,
                     desc='True if this parameter can be used as a dynamic control, else False.'
                          'If _unspecified, attempt to determine through introspection.',
                     deprecation="Option dynamic has been replaced by option 'static_target' and "
                                 "will be removed in Dymos 2.0.0.\nNote that 'static_target' has "
                                 "the opposite meaning of option 'dynamic', so parameters with "
                                 "option 'dynamic' set to False should now use 'static_target' set "
                                 "to True.")

        self.declare(name='static_target', default=_unspecified,
                     desc='True if the target of this parameter does NOT have a unique value at '
                          'each node in the ODE.'
                          'If _unspecified, attempt to determine through introspection.',
                     deprecation='Use option `static_targets` to specify whether all targets\n'
                                 'are static (static_targets=True), none are static (static_targets=False),\n'
                                 'static_targets are determined via introspection (static_targets=_unspecified),\n'
                                 'or give an explicit sequence of the static targets.')

        self.declare(name='static_targets', default=_unspecified,
                     desc='If a boolean, specifies whether all targets are static (True), or no\n'
                          'targets are static (False). Otherwise, provide a list of the static\n'
                          'targets within the ODE. If left unspecified, static targets will be\n'
                          'determined by finding inptus tagged with \'dymos.static_target\'.')

        self.declare(name='targets', allow_none=True, default=_unspecified,
                     desc='Targets in the ODE to which the state is connected')

        self.declare(name='val', types=(Iterable, np.ndarray, Number), default=0.0,
                     desc='The default value of the parameter in the phase.')

        self.declare(name='shape', check_valid=check_valid_shape, default=_unspecified, allow_none=True,
                     desc='The shape of the parameter.')

        self.declare(name='lower', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The lower bound of the parameter. This '
                          'option is invalid if opt=False.')

        self.declare(name='upper', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The upper bound of the parameter. This '
                          'option is invalid if opt=False.')

        self.declare(name='scaler', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The scaler of the parameter. This '
                          'option is invalid if opt=False.')

        self.declare(name='adder', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The adder of the parameter. This '
                          'option is invalid if opt=False.')

        self.declare(name='ref0', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The zero-reference value of the parameter. This '
                          'option is invalid if opt=False.')

        self.declare(name='ref', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='The unit-reference value of the parameter. This '
                          'option is invalid if opt=False.')

        self.declare(name='include_timeseries', types=bool, default=None,
                     allow_none=True,
                     desc='True if the static parameters should be included in output timeseries, else False.'
                          'If None (default) set the value based on Phase.timeseries_options["include_parameters"]')


class TrajParameterOptionsDictionary(ParameterOptionsDictionary):
    """
    An OptionsDictionary specific to trajectory design parameters.

    Parameters
    ----------
    read_only : bool
        If True, setting (via __setitem__ or update) is not permitted.
    """
    def __init__(self, read_only=False):
        super(TrajParameterOptionsDictionary, self).__init__(read_only)

        self.declare(name='custom_targets', types=dict, default=None, allow_none=True,
                     desc='Used to override the default targets of the trajectory input parameter'
                          ' in each phase.  By default its target will be the same as its name')

        self._dict.pop('targets')

        self.declare(name='targets', types=dict, default=None, allow_none=True,
                     desc='Used to specify the targets for the parameter in each phase. '
                          'If None, Dymos will attempt to connect it to a parameter of '
                          'the same name in each phase.  Otherwise, targets should be a given '
                          'as a dictionary.  For each phase name given as a key in the dictionary,'
                          'if the associated value is a string, connect the parameter to the phase'
                          ' input parameter given by the string. If the associated value is a'
                          ' sequence, treat it as a list of ODE-relative targets for the parameter'
                          ' in that phase')


class StateOptionsDictionary(om.OptionsDictionary):
    """
    An OptionsDictionary specific to states.

    Parameters
    ----------
    read_only : bool
        If True, setting (via __setitem__ or update) is not permitted.
    """
    def __init__(self, read_only=False):
        super(StateOptionsDictionary, self).__init__(read_only)

        self.declare(name='name', types=str,
                     desc='name of ODE state variable')

        self.declare(name='units', default=_unspecified,
                     allow_none=True, desc='units in which the state variable is defined')

        self.declare(name='opt', types=bool, default=True,
                     desc='If true, the values of this state are a design variable '
                          'for the optimizer.  Otherwise it exists as an unconnected '
                          'input.')

        self.declare(name='fix_initial', types=(bool, Iterable), default=False,
                     desc='If True, the initial value of this state is fixed and not a '
                          'design variable. If the state variable has a non-scalar shape, '
                          'this may be an iterable of bool for each index. '
                          'This option is invalid if opt=False.')

        self.declare(name='fix_final', types=(bool, Iterable), default=False,
                     desc='If True, the final value of this state is fixed and not a '
                          'design variable. If the state variable has a non-scalar shape, '
                          'this may be an iterable of bool for each index. This option is '
                          'invalid if opt=False.')

        self.declare(name='initial_bounds', types=Iterable, default=None, allow_none=True,
                     desc='Bounds on the value of the state at the start of the phase. This '
                          'option is invalid if opt=False.')

        self.declare(name='final_bounds', types=Iterable, default=None, allow_none=True,
                     desc='Bounds on the value of the state at the end of the phase. This '
                          'option is invalid if opt=False.')

        self.declare(name='val', types=(Iterable, Number), default=1.0,
                     desc='Default value of the state variable at the discretization nodes')

        self.declare(name='desc', types=str, default='',
                     desc='description of the state variable')

        self.declare(name='shape', types=Iterable, allow_none=True, default=None,
                     desc='shape of the state variable, as determined by introspection')

        self.declare(name='rate_source', types=str, allow_none=True, default=None,
                     desc='ODE-path or phase variable providing the derivative of the state variable')

        self.declare(name='source', types=str, allow_none=True, default=None,
                     desc='RHS-path or phase variable providing value of the state variable, for Analytic '
                          'transcription only!')

        self.declare(name='targets', allow_none=True, default=_unspecified,
                     desc='Targets in the ODE to which the state is connected')

        self.declare(name='lower', types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='Lower bound of the state variable at the discretization nodes. This '
                          'option is invalid if opt=False.')

        self.declare(name='upper',
                     types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='Upper bound of the state variable at the discretization nodes. This '
                          'option is invalid if opt=False.')

        self.declare(name='scaler',
                     types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='Scaler of the state variable at the discretization nodes. This '
                          'option is invalid if opt=False.')

        self.declare(name='adder',
                     types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='Adder of the state variable at the discretization nodes. This '
                          'option is invalid if opt=False.')

        self.declare(name='ref0',
                     types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='Zero-reference value of the state variable at the discretization nodes. '
                          'This option is invalid if opt=False.')

        self.declare(name='ref',
                     types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='Unit-reference value of the state variable at the discretization nodes. '
                          'This option is invalid if opt=False.')

        self.declare(name='defect_scaler',
                     types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='Scaler of the state variable defects at the collocation nodes. '
                          'If defect_scaler and defect_ref are both None but the state scaler '
                          'or ref are provided, use those values as the defect scaler or ref. '
                          'This option is invalid if opt=False.')

        self.declare(name='defect_ref',
                     types=(Iterable, Number), default=None,
                     allow_none=True,
                     desc='Unit-reference value of the state defects at the collocation nodes. This'
                          ' option is invalid if opt=False.  If provided, this option overrides'
                          ' defect_scaler. If defect_scaler and defect_ref are both None but the '
                          'state scaler or ref are provided, use those values as the defect '
                          'scaler or ref.')

        self.declare(name='continuity', types=(bool, dict), default=True,
                     desc='Enforce continuity of state values at segment boundaries. This '
                          'option is invalid if opt=False.')

        self.declare(name='continuity_scaler', types=(Number,), default=None, allow_none=True,
                     desc='Scaler for continuity at segment boundaries. This '
                          'option is invalid if opt=False.')

        self.declare(name='continuity_ref', types=(Number,), default=None, allow_none=True,
                     desc='Reference unit value for continuity at segment boundaries instead of scaler.'
                          'This option is invalid if opt=False.')

        self.declare(name='solve_segments', default=None, allow_none=True,
                     values=(False, 'forward', 'backward'),
                     desc='If \'forward\', collocation defects within each'
                          'segment are solved with a Newton solver by fixing the initial value in the'
                          'phase (if using compressed transcription) or segment (if not using '
                          'compressed transcription). This provides a forward shooting (or multiple shooting)'
                          'method.  If \'backward\', the final value in the phase or segment is fixed'
                          'and a solver finds the other ones to mimic reverse propagation. If None, '
                          '(the default) use the value of solve_segments in the transcription. Set '
                          'to False to explicitly disable the use of a solver to converge the state'
                          'time history.')

        self.declare(name='connected_initial', default=False, types=bool,
                     desc='Whether an input is created to pass in the initial state. This may be '
                          'set by a trajectory that links phases.',
                     deprecation="State option 'connected_initial' is deprecated. Please use 'input_initial' instead.")

        self.declare(name='input_initial', default=False, types=bool,
                     desc='Whether the initial value of the state is expected to be connected to an exterior value.')

        self.declare(name='input_final', default=False, types=bool,
                     desc='Whether the final value of the state is expected to be connected to an exterior value.')


class TimeOptionsDictionary(om.OptionsDictionary):
    """
    An OptionsDictionary for time options.

    Parameters
    ----------
    read_only : bool
        If True, setting (via __setitem__ or update) is not permitted.
    """
    def __init__(self, read_only=False):
        super(TimeOptionsDictionary, self).__init__(read_only)

        self.declare('name', types=str, default='time', desc='Name of the integraiton variable in the phase.')

        self.declare('units', types=str, allow_none=True,
                     default='s', desc='Units for the integration variable')

        self.declare(name='fix_initial', types=bool, default=False,
                     desc='If True, the initial value of time is not a design variable.')

        self.declare(name='fix_duration', types=bool, default=False,
                     desc='If True, the  phase duration is not a design variable.')

        self.declare(name='input_initial', types=bool, default=False,
                     desc='If True, the initial value of time (t_initial) is expected to be '
                          'connected to an external output source.')

        self.declare(name='input_duration', types=bool, default=False,
                     desc='If True, the phase duration (t_duration) is expected to be '
                          'connected to an external output source.')

        self.declare('initial_val', types=Number, default=0.0,
                     desc='Value of the integration variable at the start of the phase.')

        self.declare('initial_bounds', types=Iterable, default=(None, None),
                     desc='Tuple of (lower, upper) bounds for the integration variable at '
                          'the start of the phase.')

        self.declare('initial_scaler', types=Number, allow_none=True, default=None,
                     desc='Scalar for the initial value of the integration variable.')

        self.declare('initial_adder', types=Number, allow_none=True, default=None,
                     desc='Adder for the initial value of the integration variable.')

        self.declare('initial_ref0', types=Number, allow_none=True, default=None,
                     desc='Zero-reference value for the initial value of the integration variable.')

        self.declare('initial_ref', types=Number, allow_none=True, default=None,
                     desc='Unit-reference value for the initial value of the integration variable.')

        self.declare('duration_val', types=Number, default=1.0,
                     desc='Value of the duration of the integration variable across the phase.')

        self.declare('duration_bounds', types=Iterable, default=(None, None),
                     desc='Tuple of (lower, upper) bounds for the duration '
                          'of the integration variable across the phase.')

        self.declare('duration_scaler', types=Number, allow_none=True, default=None,
                     desc='Scalar for the duration of the integration variable across the phase.')

        self.declare('duration_adder', types=Number, allow_none=True, default=None,
                     desc='Adder for the duration of the integration variable across the phase.')

        self.declare('duration_ref0', types=Number, allow_none=True, default=None,
                     desc='Zero-reference value for the duration of the integration variable '
                          'across the phase.')

        self.declare('duration_ref', types=Number, allow_none=True, default=None,
                     desc='Unit-reference value for the duration of the integration variable '
                          'across the phase.')

        self.declare(name='targets', allow_none=True, default=_unspecified,
                     desc='targets in the ODE to which the integration variable is connected')

        self.declare(name='time_phase_targets', allow_none=True, default=_unspecified,
                     desc='targets in the ODE to which the elapsed duration of the phase is connected')

        self.declare(name='t_initial_targets', allow_none=True, default=[],
                     desc='targets in the ODE to which the initial time of the phase is connected')

        self.declare(name='t_duration_targets', allow_none=True, default=[],
                     desc='targets in the ODE to which the total duration of the phase is connected')

        self.declare(name='t_final_targets', allow_none=True, default=[],
                     desc='targets in the ODE to which the final time of the phase is connected')

        self.declare(name='t_duration_balance_options', default={},
                     desc='options dictionary for the duration residual')


class GridRefinementOptionsDictionary(om.OptionsDictionary):
    """
    An OptionsDictionary for grid refinement options in a Phase.

    Parameters
    ----------
    read_only : bool
        If True, setting (via __setitem__ or update) is not permitted.
    """
    def __init__(self, read_only=False):
        super(GridRefinementOptionsDictionary, self).__init__(read_only)

        self.declare('refine', types=bool, default=True,
                     desc='If True, this Phase may be refined during the grid refinement procedure.')

        self.declare(name='tolerance', types=float, default=1.0E-4,
                     desc='Default tolerance for grid refinement in this phase.')

        self.declare(name='min_order', types=int, default=3,
                     desc='Minimum transcription order for segments in this phase.')

        self.declare(name='max_order', types=int, default=14,
                     desc='Maximum transcription order for segments in this phase.')
        self.declare(name='smoothness_factor', types=float, default=1.2,
                     desc='Maximum allowed ratio of state second derivatives across refinement iterations')


class SimulateOptionsDictionary(om.OptionsDictionary):
    """
    An OptionsDictionary for simulate options in a Phase.

    Parameters
    ----------
    read_only : bool
        If True, setting (via __setitem__ or update) is not permitted.
    """
    def __init__(self, read_only=False):
        super(SimulateOptionsDictionary, self).__init__(read_only)

        self.declare('method', values=('RK23', 'RK45', 'DOP853', 'BDF', 'Radau', 'LSODA'), default='DOP853',
                     desc='The method used by simulate to propagate the ODE.')

        self.declare(name='atol', types=(float, np.array), default=1.0E-6,
                     desc='Absolute error tolerance for variable step integration.')

        self.declare(name='rtol', types=(float, np.array), default=1.0E-3,
                     desc='Relative error tolerance for variable step integration.')

        self.declare(name='first_step', types=float, allow_none=True, default=None,
                     desc='Initial step size, or None if the algorithm should choose.')

        self.declare(name='max_step', types=float, default=np.inf,
                     desc='Maximum allowable step size')

        self.declare(name='times_per_seg', types=int, allow_none=True, default=10,
                     desc='The default number of output times per segment for Phase.simulate.'
                          'If None, and not provided as an argument to simulate, use the'
                          'same grid as the Phase transcription.')


class ConstraintOptionsDictionary(om.OptionsDictionary):
    """
    An OptionsDictionary for path and boundary constraints.

    Parameters
    ----------
    read_only : bool
        If True, setting (via __setitem__ or update) is not permitted.
    """
    def __init__(self, read_only=False):
        super(ConstraintOptionsDictionary, self).__init__(read_only)

        self.declare(name='name', types=str, default=None, allow_none=True,
                     desc='Name or ODE-relative path of the variable to be constrained.')

        self.declare(name='constraint_name', types=str, default=None, allow_none=True,
                     desc='Name of the variable when used as a constraint, to avoid name collisions.')

        self.declare(name='constraint_path', types=str, default=None, allow_none=True,
                     desc='Path in the phase to the constrained output. Determined automatically by dymos.')

        self.declare(name='lower', types=(Iterable, Number), default=None, allow_none=True,
                     desc='Lower bound of the constraint.')

        self.declare(name='upper', types=(Iterable, Number), default=None, allow_none=True,
                     desc='Upper bound of the constraint.')

        self.declare(name='equals', types=(Iterable, Number), default=None, allow_none=True,
                     desc='Desired vlue for an equality constraint.')

        self.declare(name='scaler', types=(Iterable, Number), default=None, allow_none=True,
                     desc='Scaler of the variable.')

        self.declare(name='adder', types=(Iterable, Number), default=None, allow_none=True,
                     desc='Adder of the state variable.')

        self.declare(name='ref0', types=(Iterable, Number), default=None, allow_none=True,
                     desc='Zero-reference value of the variable.')

        self.declare(name='ref', types=(Iterable, Number), default=None, allow_none=True,
                     desc='Unit-reference value of the variable.')

        self.declare(name='indices', types=(Iterable,), default=None, allow_none=True,
                     desc='Indices value of the variable, format is controlled by the `flat_indices` option.')

        self.declare(name='shape', types=(Iterable,), default=None, allow_none=True,
                     desc='The shape of the constrained variable. This is generally determined automatically by dymos.')

        self.declare(name='linear', types=(bool,), default=False,
                     desc='If True, tell the optimizer to treat this as a linear constraint. Setting this to True '
                          'when the constraint is not actually linear will result in a failure of the optimization.')

        self.declare(name='units', types=str, default=None, allow_none=True,
                     desc='Units to be used for the constraint bounds, or None to use the units of the constrained '
                          'variable.')

        self.declare(name='flat_indices', types=bool, default=True,
                     desc='If True, the given indices will be treated as indices into a C-order flattened array based '
                          'on the shaped of the constrained variable at a point in time.')


class TimeseriesOutputOptionsDictionary(om.OptionsDictionary):
    """
    An OptionsDictionary for timeseries outputs.

    Parameters
    ----------
    read_only : bool
        If True, setting (via __setitem__ or update) is not permitted.
    """
    def __init__(self, read_only=False):
        super(TimeseriesOutputOptionsDictionary, self).__init__(read_only)

        self.declare(name='name', types=str, default=None, allow_none=True,
                     desc='Name or ODE-relative path of the variable to be output in the timeseries.')

        self.declare(name='output_name', types=str, default=None, allow_none=True,
                     desc='Name of the variable used as the output from the timeseries, to avoid name collisions.')

        self.declare(name='wildcard_units', types=dict, default=None, allow_none=True,
                     desc='Variable name, unit mapping that can be provided if timeseries are specified '
                          'with wildcards.')

        self.declare(name='shape', default=None, allow_none=True,
                     desc='The shape of the timeseries output variable.'
                          ' This is generally determined automatically by dymos.')

        self.declare(name='units', default=None, allow_none=True,
                     desc='Units to be used for the timeseries output, or None to leave the units unchanged.')

        self.declare(name='src', types=str, default=None, allow_none=True,
                     desc='The phase-relative path of the source of this timeseries output,'
                          ' used when issuing connections.')

        self.declare(name='src_idxs', types=(Iterable,), default=None, allow_none=True,
                     desc='The indices of the source of this timeseries output to be used when issuing connections.')

        self.declare(name='val', types=bool, default=True,
                     desc='If True, include the value of this variable as a timeseries output.')

        self.declare(name='is_rate', default=False, allow_none=False,
                     desc='If True this is a rate.')


class PhaseTimeseriesOptionsDictionary(om.OptionsDictionary):
    """
    An OptionsDictionary for phase options related to timeseries.

    Parameters
    ----------
    read_only : bool
        If True, setting (via __setitem__ or update) is not permitted.
    """
    def __init__(self, read_only=False):
        super().__init__(read_only)

        self.declare(name='use_prefix', types=bool, default=False,
                     desc='If True, prefix the timeseries variable output with the type '
                          'of variable (this is legacy behavior that changed in Dymos 1.8.0)')

        self.declare(name='include_state_rates', types=bool, default=False,
                     desc='If True, include state rates in the timeseries outputs by default.')

        self.declare(name='include_control_rates', types=bool, default=False,
                     desc='If True, include control rates in the timeseries outputs by default.')

        self.declare(name='include_t_phase', types=bool, default=False,
                     desc='If True, include the elapsed phase time in the timeseries outputs by default.')

        self.declare(name='include_parameters', types=bool, default=False,
                     desc='If True, include the parameters in the timeseries outputs by default.')



import unittest
import itertools

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.transcriptions.grid_data import GridData
from dymos.transcriptions.common import GaussLobattoContinuityComp, RadauPSContinuityComp
from dymos.phase.options import StateOptionsDictionary, ControlOptionsDictionary
from dymos.utils.testing_utils import assert_check_partials

# Modify class so we can run it standalone.
from dymos.utils.misc import CompWrapperConfig
GaussLobattoContinuityComp = CompWrapperConfig(GaussLobattoContinuityComp)
RadauPSContinuityComp = CompWrapperConfig(RadauPSContinuityComp)

params_list = itertools.product(['gauss-lobatto', 'radau-ps'],  # transcription
                                ['compressed', 'uncompressed']  # compressed
                                )


@use_tempdirs
class TestContinuityComp(unittest.TestCase):

    def test_continuity_comp(self):
        for transcription, compressed in params_list:
            with self.subTest():

                dm.options['include_check_partials'] = True
                num_seg = 3

                gd = GridData(num_segments=num_seg,
                              transcription_order=[5, 3, 3],
                              segment_ends=[0.0, 3.0, 10.0, 20],
                              transcription=transcription,
                              compressed=compressed == 'compressed')

                self.p = om.Problem(model=om.Group())

                ivp = self.p.model.add_subsystem('ivc', subsys=om.IndepVarComp(), promotes_outputs=['*'])

                nn = gd.subset_num_nodes['all']

                ivp.add_output('x', val=np.arange(nn), units='m')
                ivp.add_output('y', val=np.arange(nn), units='m/s')
                ivp.add_output('u', val=np.zeros((nn, 3)), units='deg')
                ivp.add_output('v', val=np.arange(nn), units='N')
                ivp.add_output('u_rate', val=np.zeros((nn, 3)), units='deg/s')
                ivp.add_output('v_rate', val=np.arange(nn), units='N/s')
                ivp.add_output('u_rate2', val=np.zeros((nn, 3)), units='deg/s**2')
                ivp.add_output('v_rate2', val=np.arange(nn), units='N/s**2')
                ivp.add_output('t_duration', val=121.0, units='s')

                self.p.model.add_design_var('x', lower=0, upper=100)

                state_options = {'x': StateOptionsDictionary(),
                                 'y': StateOptionsDictionary()}
                control_options = {'u': ControlOptionsDictionary(),
                                   'v': ControlOptionsDictionary()}

                state_options['x']['units'] = 'm'
                state_options['y']['units'] = 'm/s'

                # Need these for later in the test.
                state_options['x']['shape'] = (1,)
                state_options['y']['shape'] = (1,)

                control_options['u']['units'] = 'deg'
                control_options['u']['shape'] = (3,)
                control_options['u']['continuity'] = True

                control_options['v']['units'] = 'N'

                # Need these for later in the test.
                state_options['x']['shape'] = (1,)
                state_options['y']['shape'] = (1,)
                control_options['v']['shape'] = (1,)

                control_options['u']['rate_continuity'] = True
                control_options['v']['rate_continuity'] = True

                control_options['u']['rate2_continuity'] = True
                control_options['v']['rate2_continuity'] = True

                if transcription == 'gauss-lobatto':
                    cnty_comp = GaussLobattoContinuityComp(grid_data=gd, time_units='s',
                                                           state_options=state_options,
                                                           control_options=control_options)
                elif transcription == 'radau-ps':
                    cnty_comp = RadauPSContinuityComp(grid_data=gd, time_units='s',
                                                      state_options=state_options,
                                                      control_options=control_options)
                else:
                    raise ValueError('unrecognized transcription')

                self.p.model.add_subsystem('cnty_comp', subsys=cnty_comp)
                # The sub-indices of state_disc indices that are segment ends
                num_seg_ends = gd.subset_num_nodes['segment_ends']
                segment_end_idxs = np.reshape(gd.subset_node_indices['segment_ends'],
                                              (num_seg_ends, 1))

                if compressed != 'compressed':
                    self.p.model.connect('x', 'cnty_comp.states:x', src_indices=(segment_end_idxs,))
                    self.p.model.connect('y', 'cnty_comp.states:y', src_indices=(segment_end_idxs,))

                self.p.model.connect('t_duration', 'cnty_comp.t_duration')

                size_u = nn * np.prod(control_options['u']['shape'])
                src_idxs_u = np.arange(size_u).reshape((nn,) + control_options['u']['shape'])
                src_idxs_u = src_idxs_u[gd.subset_node_indices['segment_ends'], ...]

                size_v = nn * np.prod(control_options['v']['shape'])
                src_idxs_v = np.arange(size_v).reshape((nn,) + control_options['v']['shape'])
                src_idxs_v = src_idxs_v[gd.subset_node_indices['segment_ends'], ...]

                src_idxs_u = (src_idxs_u,)
                src_idxs_v = (src_idxs_v,)

                # if transcription =='radau-ps' or compressed != 'compressed':
                self.p.model.connect('u', 'cnty_comp.controls:u', src_indices=src_idxs_u,
                                     flat_src_indices=True)

                self.p.model.connect('u_rate', 'cnty_comp.control_rates:u_rate', src_indices=src_idxs_u,
                                     flat_src_indices=True)

                self.p.model.connect('u_rate2', 'cnty_comp.control_rates:u_rate2', src_indices=src_idxs_u,
                                     flat_src_indices=True)

                # if transcription =='radau-ps' or compressed != 'compressed':
                self.p.model.connect('v', 'cnty_comp.controls:v', src_indices=src_idxs_v,
                                     flat_src_indices=True)

                self.p.model.connect('v_rate', 'cnty_comp.control_rates:v_rate', src_indices=src_idxs_v,
                                     flat_src_indices=True)

                self.p.model.connect('v_rate2', 'cnty_comp.control_rates:v_rate2', src_indices=src_idxs_v,
                                     flat_src_indices=True)

                self.p.setup(check=True, force_alloc_complex=True)

                self.p['x'] = np.random.rand(*self.p['x'].shape)
                self.p['y'] = np.random.rand(*self.p['y'].shape)
                self.p['u'] = np.random.rand(*self.p['u'].shape)
                self.p['v'] = np.random.rand(*self.p['v'].shape)
                self.p['u_rate'] = np.random.rand(*self.p['u'].shape)
                self.p['v_rate'] = np.random.rand(*self.p['v'].shape)
                self.p['u_rate2'] = np.random.rand(*self.p['u'].shape)
                self.p['v_rate2'] = np.random.rand(*self.p['v'].shape)

                self.p.run_model()

                if compressed != 'compressed':
                    for state in ('x', 'y'):
                        xpectd = self.p[state][segment_end_idxs, ...][2::2, ...] - \
                            self.p[state][segment_end_idxs, ...][1:-1:2, ...]

                        assert_near_equal(self.p['cnty_comp.defect_states:{0}'.format(state)],
                                          xpectd.reshape((num_seg - 1,) + state_options[state]['shape']))

                for ctrl in ('u', 'v'):

                    xpectd = self.p[ctrl][segment_end_idxs, ...][2::2, ...] - \
                        self.p[ctrl][segment_end_idxs, ...][1:-1:2, ...]

                    if compressed != 'compressed':
                        assert_near_equal(self.p['cnty_comp.defect_controls:{0}'.format(ctrl)],
                                          xpectd.reshape((num_seg - 1,) + control_options[ctrl]['shape']))

                np.set_printoptions(linewidth=1024)
                cpd = self.p.check_partials(method='cs', out_stream=None)
                assert_check_partials(cpd)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

from dymos.utils.misc import get_rate_units


class TestMisc(unittest.TestCase):

    def test_get_rate_units(self):
        rate_units = get_rate_units('m', 's', deriv=1)

        self.assertEqual(rate_units, 'm/s')

        rate2_units = get_rate_units('m', 's', deriv=2)

        self.assertEqual(rate2_units, 'm/s**2')

    def test_get_rate_units_invalid_deriv(self):

        with self.assertRaises(ValueError) as e:
            get_rate_units('m', 's', deriv=0)
        self.assertEqual(str(e.exception), 'deriv argument must be 1 or 2.')


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np
import scipy
import scipy.sparse as sp

import openmdao.api as om

from dymos.transcriptions.grid_data import GridData
from dymos.utils.misc import get_rate_units
from dymos._options import options as dymos_options
from dymos.utils.lgl import lgl
from dymos.utils.cgl import cgl
from dymos.utils.birkhoff import birkhoff_matrix


class BirkhoffDefectComp(om.ExplicitComponent):
    """
    Class definition for the BirkhoffDefectComp.

    BirkhoffDefectComp computes the generalized defects of a segment for implicit collocation.
    There are four defects to be evaluated; state, state rate, initial state, and final state.
    The state defect is the difference between the state value design variables and the state rate
    design variables.
    The state rate defect is the difference between the state rate design variables and the computed
    state derivative at the collocation nodes.
    The initial and final state defects are the differences between the initial and final state
    design variables and the first and last index of the state design variable array.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare(
            'grid_data', types=GridData,
            desc='Container object for grid info')

        self.options.declare(
            'state_options', types=dict,
            desc='Dictionary of state names/options for the phase')

        self.options.declare(
            'time_units', default=None, allow_none=True, types=str,
            desc='Units of time')

    def configure_io(self, phase):
        """
        I/O creation is delayed until configure so we can determine shape and units.

        Parameters
        ----------
        phase : Phase
            The phase object that contains this collocation comp.
        """
        gd = self.options['grid_data']
        num_nodes = gd.subset_num_nodes['col']
        num_segs = gd.num_segments
        time_units = self.options['time_units']
        state_options = self.options['state_options']

        self.add_input('dt_dstau', units=time_units, shape=(gd.subset_num_nodes['col'],))
        self.var_names = var_names = {}
        for state_name, options in state_options.items():
            var_names[state_name] = {
                'f_value': f'state_rates:{state_name}',
                'f_computed': f'f_computed:{state_name}',
                'state_value': f'states:{state_name}',
                'state_initial_value': f'initial_states:{state_name}',
                'state_final_value': f'final_states:{state_name}',
                'state_defect': f'state_defects:{state_name}',
                'state_rate_defect': f'state_rate_defects:{state_name}',
                'initial_state_defect': f'initial_state_defects:{state_name}',
                'final_state_defect': f'final_state_defects:{state_name}',
                'state_continuity_defect': f'state_cnty_defects:{state_name}'
            }

        for state_name, options in state_options.items():
            shape = options['shape']
            units = options['units']

            rate_units = get_rate_units(units, time_units)
            var_names = self.var_names[state_name]

            rate_source = options['rate_source']
            rate_source_type = phase.classify_var(rate_source)

            self.add_input(
                name=var_names['f_value'],
                shape=(num_nodes,) + shape,
                desc=f'Estimated derivative of state {state_name} at the polynomial nodes',
                units=units)

            if rate_source_type == 'state':
                var_names['f_computed'] = f'states:{rate_source}'
            else:
                self.add_input(
                    name=var_names['f_computed'],
                    shape=(num_nodes,) + shape,
                    desc=f'Computed derivative of state {state_name} at the polynomial nodes',
                    units=rate_units)

            self.add_input(
                name=var_names['state_value'],
                shape=(num_nodes,) + shape,
                desc=f'Value of the state {state_name} at the polynomial nodes',
                units=units
            )

            self.add_input(
                name=var_names['state_initial_value'],
                shape=(1,) + shape,
                desc=f'Desired initial value of state {state_name}',
                units=units
            )

            self.add_input(
                name=var_names['state_final_value'],
                shape=(1,) + shape,
                desc=f'Desired final value of state {state_name}',
                units=units
            )

            self.add_output(
                name=var_names['state_defect'],
                shape=(num_nodes + num_segs,) + shape,
                units=units
            )

            self.add_output(
                name=var_names['state_rate_defect'],
                shape=(num_nodes,) + shape,
                units=units
            )

            if 'defect_ref' in options and options['defect_ref'] is not None:
                defect_ref = options['defect_ref']
            elif 'defect_scaler' in options and options['defect_scaler'] is not None:
                defect_ref = np.divide(1.0, options['defect_scaler'])
            else:
                if 'ref' in options and options['ref'] is not None:
                    defect_ref = options['ref']
                elif 'scaler' in options and options['scaler'] is not None:
                    defect_ref = np.divide(1.0, options['scaler'])
                else:
                    defect_ref = 1.0

            if not np.isscalar(defect_ref):
                defect_ref = np.asarray(defect_ref)
                if defect_ref.shape == shape:
                    defect_ref_state = np.tile(defect_ref.flatten(), num_nodes+num_segs)
                    defect_ref_v = np.tile(defect_ref.flatten(), num_nodes)
                else:
                    raise ValueError('array-valued scaler/ref must length equal to state-size')
            else:
                defect_ref_state = defect_ref
                defect_ref_v = defect_ref

            if not options['solve_segments']:
                self.add_constraint(name=var_names['state_defect'],
                                    equals=0.0,
                                    ref=defect_ref_state,
                                    linear=True)

                self.add_constraint(name=var_names['state_rate_defect'],
                                    equals=0.0,
                                    ref=defect_ref_v)

        A_blocks = []
        B_blocks = []
        C_blocks = []

        # _xv_idxs is a set of indices that arranges the stacked
        # [[X^T],[V^T]] arrays into a segment-by-segment ordering
        # instead of all of the state values followed by all of the state rates.
        self._xv_idxs = []

        idx0 = 0

        for i in range(num_segs):
            nnps_i = gd.subset_num_nodes_per_segment['all'][0]
            if gd.grid_type == 'lgl':
                tau_i, w_i = lgl(nnps_i)
            elif gd.grid_type == 'cgl':
                tau_i, w_i = cgl(nnps_i)
            else:
                raise ValueError('invalid grid type')

            B_i = birkhoff_matrix(tau_i, w_i, grid_type=gd.grid_type)

            A_i = np.zeros((nnps_i + 1, 2 * nnps_i))
            A_i[:nnps_i, :nnps_i] = np.eye(nnps_i)
            A_i[:nnps_i, nnps_i:] = -B_i
            A_i[-1, nnps_i:] = B_i[-1, :]

            C_i = np.zeros((nnps_i + 1, 2))
            C_i[:-1, 0] = 1.
            C_i[-1, :] = [-1, 1]

            A_blocks.append(A_i)
            B_blocks.append(B_i)
            C_blocks.append(C_i)

            ar_nnps = np.arange(nnps_i, dtype=int)
            self._xv_idxs.extend(idx0 + ar_nnps)
            self._xv_idxs.extend(idx0 + ar_nnps + num_nodes)
            idx0 += nnps_i

        self._A = scipy.linalg.block_diag(*A_blocks)
        self._B = scipy.linalg.block_diag(*B_blocks)
        self._C = scipy.linalg.block_diag(*C_blocks)

        # Setup partials
        for state_name, options in state_options.items():
            shape = options['shape']
            units = options['units']
            size = np.prod(shape)

            ar1 = np.arange(num_nodes * size)
            c1 = np.repeat(np.arange(num_nodes), size)

            var_names = self.var_names[state_name]

            # The derivative of the state defect wrt x_ab is [-C].
            # The derivative of x_ab wrt x_a is eye(size).
            # Take the Kronecker product of that and an identity matrix
            # of the state's size to get the pattern that includes each
            # individual element in the state.

            # The derivative of x_ab wrt x_a is [I(size), 0(size)]^T
            # The derivative of x_ab wrt x_b is [0(size), I(size)]^T
            d_state_defect_dxa = sp.kron(np.ones((self._C.shape[0], 1)), -sp.eye(size), format='coo')
            d_state_defect_dxa.data[-size:] = 1.0
            d_dxa_r, d_dxa_c = d_state_defect_dxa.nonzero()
            d_dxa_data = d_state_defect_dxa.data.ravel()

            self.declare_partials(of=var_names['state_defect'],
                                  wrt=var_names['state_initial_value'],
                                  rows=d_dxa_r, cols=d_dxa_c, val=d_dxa_data)

            self.declare_partials(of=var_names['state_defect'],
                                  wrt=var_names['state_final_value'],
                                  rows=d_dxa_r[-size:], cols=d_dxa_c[-size:], val=-1.0)

            d_state_defect_dXV = self._A
            dXV_dX = np.vstack((np.eye(num_nodes), np.zeros((num_nodes, num_nodes))))[self._xv_idxs]
            dXV_dV = np.vstack((np.zeros((num_nodes, num_nodes)), np.eye(num_nodes)))[self._xv_idxs]
            d_state_defect_dX = sp.csr_matrix(np.kron(np.dot(d_state_defect_dXV, dXV_dX), np.eye(size)))
            d_state_defect_dV = sp.csr_matrix(np.kron(np.dot(d_state_defect_dXV, dXV_dV), np.eye(size)))

            rs_dX, cs_dX = d_state_defect_dX.nonzero()
            rs_dV, cs_dV = d_state_defect_dV.nonzero()

            self.declare_partials(of=var_names['state_defect'],
                                  wrt=var_names['state_value'],
                                  rows=rs_dX, cols=cs_dX, val=d_state_defect_dX.data.ravel())

            self.declare_partials(of=var_names['state_defect'],
                                  wrt=var_names['f_value'],
                                  rows=rs_dV, cols=cs_dV, val=d_state_defect_dV.data.ravel())

            self.declare_partials(of=var_names['state_rate_defect'],
                                  wrt=var_names['f_value'],
                                  rows=ar1, cols=ar1, val=1.0)

            self.declare_partials(of=var_names['state_rate_defect'],
                                  wrt=var_names['f_computed'],
                                  rows=ar1, cols=ar1)

            self.declare_partials(of=var_names['state_rate_defect'],
                                  wrt='dt_dstau',
                                  rows=ar1, cols=c1)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        outputs : Vector
            Unscaled, dimensional output variables read via outputs[key].
        discrete_inputs : dict or None
            If not None, dict containing discrete input values.
        discrete_outputs : dict or None
            If not None, dict containing discrete output values.
        """
        dt_dstau = np.atleast_2d(inputs['dt_dstau']).T
        num_segs = self.options['grid_data'].num_segments

        for state_name, options in self.options['state_options'].items():
            var_names = self.var_names[state_name]

            x_a = inputs[var_names['state_initial_value']]
            x_b = inputs[var_names['state_final_value']]
            # x_ab = inputs[var_names['state_segment_ends']]
            X = inputs[var_names['state_value']]
            V = inputs[var_names['f_value']]
            f = inputs[var_names['f_computed']]
            shape = x_a.shape[1:]
            size = x_a.size

            # Get stacked XV, but sorted in segment-by-segment order
            XV = np.vstack((X, V))[self._xv_idxs]

            x_ab = np.stack([x_a, x_b], axis=0).reshape((2,) + shape)

            state_defect = np.einsum('ij,jk...->ik...', self._A, XV) - \
                np.einsum('ij, j...->i...', self._C, x_ab)

            outputs[var_names['state_defect']] = state_defect
            outputs[var_names['state_rate_defect']] = (V - np.einsum('i...,i...->i...', f, dt_dstau))

            if num_segs > 1:
                (num_segs - 1) * size
                outputs[var_names['state_continuity_defect']] = x_ab[:-1, 1, ...] - x_ab[1:, 0, ...]

    def compute_partials(self, inputs, partials):
        """
        Compute sub-jacobian parts. The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        """
        dt_dstau = inputs['dt_dstau']

        for state_name, options in self.options['state_options'].items():
            var_names = self.var_names[state_name]
            shape = options['shape']
            size = np.prod(shape)
            f = inputs[var_names['f_computed']]

            partials[var_names['state_rate_defect'], var_names['f_computed']] = np.repeat(-dt_dstau, size)
            partials[var_names['state_rate_defect'], 'dt_dstau'] = -f.ravel()



import os
import unittest

from openmdao.utils.testing_utils import use_tempdirs
import openmdao.api as om
import dymos as dm
from dymos.utils.misc import om_version


def setup_problem(trans=dm.GaussLobatto(num_segments=10), polynomial_control=False,
                  fix_final_state=True, fix_final_control=False):
    from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

    p = om.Problem(model=om.Group())
    p.driver = om.ScipyOptimizeDriver()

    phase = dm.Phase(ode_class=BrachistochroneODE, transcription=trans)

    p.model.add_subsystem('phase0', phase)

    phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

    phase.add_state('x', fix_initial=True, fix_final=fix_final_state)
    phase.add_state('y', fix_initial=True, fix_final=fix_final_state)
    phase.add_state('v', fix_initial=True)

    if not polynomial_control:
        phase.add_control('theta', units='deg',
                          rate_continuity=False, lower=0.01, upper=179.9, fix_final=fix_final_control)
    else:
        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9,
                          fix_final=fix_final_control, control_type='polynomial')

    phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

    # Minimize time at the end of the phase
    phase.add_objective('time', loc='final', scaler=10)

    p.model.linear_solver = om.DirectSolver()

    p.setup()

    p['phase0.t_initial'] = 0.0
    p['phase0.t_duration'] = 2.0

    p['phase0.states:x'] = phase.interp('x', [0, 10])
    p['phase0.states:y'] = phase.interp('y', [10, 5])
    p['phase0.states:v'] = phase.interp('v', [0, 9.9])

    if polynomial_control:
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100.5])
    else:
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100.5])

    return p


@use_tempdirs
class TestLoadCase(unittest.TestCase):

    def test_load_case_unchanged_grid(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm

        p = setup_problem(dm.GaussLobatto(num_segments=10))

        # Solve for the optimal trajectory
        dm.run_problem(p)

        # Load the solution
        if om_version()[0] > (3, 34, 2):
            sol_file = p.get_outputs_dir() / 'dymos_solution.db'
        else:
            sol_file = 'dymos_solution.db'

        case = om.CaseReader(sol_file).get_case('final')

        # Initialize the system with values from the case.
        # We unnecessarily call setup again just to make sure we obliterate the previous solution
        p.setup()

        p.set_val('phase0.controls:theta', 0.0)

        # Load the values from the previous solution
        p.load_case(case)

        # Run the model to ensure we find the same output values as those that we recorded
        p.run_model()

        assert_near_equal(case.get_val('phase0.controls:theta'),
                          p.get_val('phase0.controls:theta'))

    def test_load_case_unchanged_grid_polynomial_control(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm

        p = setup_problem(dm.GaussLobatto(num_segments=10), polynomial_control=True)

        # Solve for the optimal trajectory
        dm.run_problem(p)

        # Load the solution
        if om_version()[0] > (3, 34, 2):
            sol_file = p.get_outputs_dir() / 'dymos_solution.db'
        else:
            sol_file = 'dymos_solution.db'

        case = om.CaseReader(sol_file).get_case('final')

        # Initialize the system with values from the case.
        # We unnecessarily call setup again just to make sure we obliterate the previous solution
        p.setup()

        # Load the values from the previous solution
        p.load_case(case)

        # Run the model to ensure we find the same output values as those that we recorded
        p.run_model()

        assert_near_equal(p.get_val('phase0.controls:theta'),
                          case.get_val('phase0.controls:theta'))

    def test_load_case_lgl_to_radau(self):
        import numpy as np
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm

        p = setup_problem(dm.GaussLobatto(num_segments=10))

        # Solve for the optimal trajectory
        dm.run_problem(p)

        # Load the solution
        if om_version()[0] > (3, 34, 2):
            sol_file = p.get_outputs_dir() / 'dymos_solution.db'
        else:
            sol_file = 'dymos_solution.db'

        case = om.CaseReader(sol_file).get_case('final')

        # create a problem with a different transcription with a different number of variables
        q = setup_problem(dm.Radau(num_segments=20))

        # Load the values from the previous solution
        q.load_case(case)

        # Run the model to ensure we find the same output values as those that we recorded
        q.run_driver()

        outputs = dict([(o[0], o[1]) for o in case.list_outputs(units=True, shape=True,
                                                                out_stream=None)])

        time_val = outputs['phase0.timeseries.time']['val']
        theta_val = outputs['phase0.timeseries.theta']['val']

        time_val_uniq, idx = np.unique(time_val, return_index=True)
        theta_val_uniq = theta_val[idx]

        assert_near_equal(q['phase0.timeseries.theta'],
                          q.model.phase0.interp(xs=time_val_uniq, ys=theta_val_uniq, nodes='all'),
                          tolerance=1.0E-3)

    def test_load_case_radau_to_lgl(self):
        import numpy as np
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm

        p = setup_problem(dm.Radau(num_segments=20))

        # Solve for the optimal trajectory
        dm.run_problem(p)

        # Load the solution
        if om_version()[0] > (3, 34, 2):
            sol_file = p.get_outputs_dir() / 'dymos_solution.db'
        else:
            sol_file = 'dymos_solution.db'

        case = om.CaseReader(sol_file).get_case('final')

        # create a problem with a different transcription with a different number of variables
        q = setup_problem(dm.GaussLobatto(num_segments=50))

        # Load the values from the previous solution
        q.load_case(case)

        # Run the model to ensure we find the same output values as those that we recorded
        q.run_model()

        time_p = case.get_val('phase0.timeseries.time')
        theta_p = case.get_val('phase0.timeseries.theta')

        time_q = q.get_val('phase0.timeseries.time')
        theta_q = q.get_val('phase0.timeseries.theta')

        time_p_unique, p_idx = np.unique(time_p, return_index=True)
        theta_p_unique = theta_p[p_idx]

        time_q_unique, q_idx = np.unique(time_q, return_index=True)
        theta_q_unique = theta_q[q_idx]

        assert_near_equal(q.model.phase0.interp(xs=time_p_unique, ys=theta_p_unique, nodes='all'),
                          q.model.phase0.interp(xs=time_q_unique, ys=theta_q_unique, nodes='all'),
                          tolerance=1.0E-2)

    def test_load_case_radau_to_birkhoff(self):
        import numpy as np
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm

        p = setup_problem(dm.Radau(num_segments=20))

        # Solve for the optimal trajectory
        dm.run_problem(p)

        # Load the solution
        if om_version()[0] > (3, 34, 2):
            sol_file = p.get_outputs_dir() / 'dymos_solution.db'
        else:
            sol_file = 'dymos_solution.db'

        case = om.CaseReader(sol_file).get_case('final')

        # create a problem with a different transcription with a different number of variables
        q = setup_problem(dm.Birkhoff(num_nodes=50))

        # Fill q with junk so that we can be sure load_case worked
        q['phase0.t_initial'] = -88
        q['phase0.t_duration'] = 88

        q['phase0.states:x'] = -88
        q['phase0.states:y'] = -88
        q['phase0.states:v'] = -88

        q['phase0.initial_states:x'] = -88
        q['phase0.initial_states:y'] = -88
        q['phase0.initial_states:v'] = -88

        q['phase0.final_states:x'] = -88
        q['phase0.final_states:y'] = -88
        q['phase0.final_states:v'] = -88

        # Load the values from the previous solution
        q.load_case(case)

        # Run the model to ensure we find the same output values as those that we recorded
        q.run_model()

        time_p = case.get_val('phase0.timeseries.time')
        theta_p = case.get_val('phase0.timeseries.theta')

        time_q = q.get_val('phase0.timeseries.time')
        theta_q = q.get_val('phase0.timeseries.theta')

        x_p = case.get_val('phase0.timeseries.x')
        y_p = case.get_val('phase0.timeseries.y')
        v_p = case.get_val('phase0.timeseries.v')

        x0_q = q.get_val('phase0.initial_states:x')
        xf_q = q.get_val('phase0.final_states:x')

        y0_q = q.get_val('phase0.initial_states:y')
        yf_q = q.get_val('phase0.final_states:y')

        v0_q = q.get_val('phase0.initial_states:v')
        vf_q = q.get_val('phase0.final_states:v')

        time_p_unique, p_idx = np.unique(time_p, return_index=True)
        theta_p_unique = theta_p[p_idx]

        time_q_unique, q_idx = np.unique(time_q, return_index=True)
        theta_q_unique = theta_q[q_idx]

        assert_near_equal(q.model.phase0.interp(xs=time_p_unique, ys=theta_p_unique, nodes='all'),
                          q.model.phase0.interp(xs=time_q_unique, ys=theta_q_unique, nodes='all'),
                          tolerance=1.0E-2)

        assert_near_equal(x_p[0, ...], x0_q, tolerance=1.0E-5)
        assert_near_equal(x_p[-1, ...], xf_q, tolerance=1.0E-5)

        assert_near_equal(y_p[0, ...], y0_q, tolerance=1.0E-5)
        assert_near_equal(y_p[-1, ...], yf_q, tolerance=1.0E-5)

        assert_near_equal(v_p[0, ...], v0_q, tolerance=1.0E-5)
        assert_near_equal(v_p[-1, ...], vf_q, tolerance=1.0E-5)

    def test_load_case_warn_fix_final_states(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_warnings
        import dymos as dm

        p = setup_problem(dm.Radau(num_segments=20))

        # Solve for the optimal trajectory
        dm.run_problem(p)

        # Load the solution
        if om_version()[0] > (3, 34, 2):
            sol_file = p.get_outputs_dir() / 'dymos_solution.db'
        else:
            sol_file = 'dymos_solution.db'

        case = om.CaseReader(sol_file).get_case('final')

        # create a problem with a different transcription with a different number of variables
        q = setup_problem(dm.GaussLobatto(num_segments=50))

        msgs = []

        # Load the values from the previous solution
        for state_name in ['x', 'y']:
            msgs.append((UserWarning, f"phase0.states:{state_name} specifies 'fix_final=True'."
                                      f" If the given restart file has a different final value"
                                      f" this will overwrite the user-specified value"))

        with assert_warnings(msgs):
            q.load_case(case)

    def test_load_case_warn_fix_final_control(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_warning
        import dymos as dm
        p = setup_problem(dm.Radau(num_segments=10))

        # Solve for the optimal trajectory
        dm.run_problem(p)

        # Load the solution
        if om_version()[0] > (3, 34, 2):
            sol_file = p.get_outputs_dir() / 'dymos_solution.db'
        else:
            sol_file = 'dymos_solution.db'

        case = om.CaseReader(sol_file).get_case('final')

        # create a problem with a different transcription with a different number of variables
        q = setup_problem(dm.Radau(num_segments=10), fix_final_state=False, fix_final_control=True)

        msg = "phase0.controls:theta specifies 'fix_final=True'. If the given restart file has a" \
              " different final value this will overwrite the user-specified value"

        with assert_warning(UserWarning, msg):
            q.load_case(case)

    def test_load_case_warn_fix_final_polynomial_control(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_warning
        import dymos as dm
        p = setup_problem(dm.Radau(num_segments=10), polynomial_control=True,)

        # Solve for the optimal trajectory
        dm.run_problem(p)

        # Load the solution
        if om_version()[0] > (3, 34, 2):
            sol_file = p.get_outputs_dir() / 'dymos_solution.db'
        else:
            sol_file = 'dymos_solution.db'

        case = om.CaseReader(sol_file).get_case('final')

        # create a problem with a different transcription with a different number of variables
        q = setup_problem(dm.Radau(num_segments=10), polynomial_control=True,
                          fix_final_state=False, fix_final_control=True)

        # Load the values from the previous solution
        msg = "phase0.controls:theta specifies 'fix_final=True'. If the given restart file has a" \
              " different final value this will overwrite the user-specified value"

        with assert_warning(UserWarning, msg):
            q.load_case(case)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



"""Test Dymos Linkage report GUI with using Playwright."""
import json
import pathlib
import unittest

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
import dymos as dm
from dymos.examples.balanced_field.balanced_field_ode import BalancedFieldODEComp

from dymos.visualization.linkage.report import _create_model_data


def _flatten_dict(d, parent_key='', sep='.'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(_flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


@use_tempdirs
class TestLinkageReport(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    def test_model_data(self):
        """
        Test that model data for the linkage report contains the expected data.
        """
        p = om.Problem()

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'IPOPT'
        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['derivative_test'] = 'first-order'

        p.driver.declare_coloring()

        # First Phase: Brake release to V1 - both engines operable
        br_to_v1 = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                            ode_init_kwargs={'mode': 'runway'})
        br_to_v1.set_time_options(fix_initial=True, duration_bounds=(1, 1000), duration_ref=10.0)
        br_to_v1.add_state('r', fix_initial=True, lower=0, ref=1000.0, defect_ref=1000.0)
        br_to_v1.add_state('v', fix_initial=True, lower=0.0001, ref=100.0, defect_ref=100.0)
        br_to_v1.add_parameter('alpha', val=0.0, opt=False, units='deg')
        br_to_v1.add_timeseries_output('*')

        # Second Phase: Rejected takeoff at V1 - no engines operable
        rto = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                       ode_init_kwargs={'mode': 'runway'})
        rto.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
        rto.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        rto.add_state('v', fix_initial=False, lower=0.0001, ref=100.0, defect_ref=100.0)
        rto.add_parameter('alpha', val=0.0, opt=False, units='deg')
        rto.add_timeseries_output('*')

        # Third Phase: V1 to Vr - single engine operable
        v1_to_vr = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                            ode_init_kwargs={'mode': 'runway'})
        v1_to_vr.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
        v1_to_vr.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        v1_to_vr.add_state('v', fix_initial=False, lower=0.0001, ref=100.0, defect_ref=100.0)
        v1_to_vr.add_parameter('alpha', val=0.0, opt=False, units='deg')
        v1_to_vr.add_timeseries_output('*')

        # Fourth Phase: Rotate - single engine operable
        rotate = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                          ode_init_kwargs={'mode': 'runway'})
        rotate.set_time_options(fix_initial=False, duration_bounds=(1.0, 5), duration_ref=1.0)
        rotate.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        rotate.add_state('v', fix_initial=False, lower=0.0001, ref=100.0, defect_ref=100.0)
        rotate.add_control('alpha', order=1, opt=True, units='deg', lower=0, upper=10,
                           ref=10, val=[0, 10], control_type='polynomial')
        rotate.add_timeseries_output('*')

        # Fifth Phase: Climb to target speed and altitude at end of runway.
        climb = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=5),
                         ode_init_kwargs={'mode': 'climb'})
        climb.set_time_options(fix_initial=False, duration_bounds=(1, 100), duration_ref=1.0)
        climb.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        climb.add_state('h', fix_initial=True, lower=0.0, ref=1.0, defect_ref=1.0)
        climb.add_state('v', fix_initial=False, lower=0.0001, ref=100.0, defect_ref=100.0)
        climb.add_state('gam', fix_initial=True, lower=0.0, ref=0.05, defect_ref=0.05)
        climb.add_control('alpha', opt=True, units='deg', lower=-10, upper=15, ref=10)
        climb.add_timeseries_output('*')

        # Instantiate the trajectory and add phases
        traj = dm.Trajectory()
        p.model.add_subsystem('traj', traj)
        traj.add_phase('br_to_v1', br_to_v1)
        traj.add_phase('rto', rto)
        traj.add_phase('v1_to_vr', v1_to_vr)
        traj.add_phase('rotate', rotate)
        traj.add_phase('climb', climb)

        # Add parameters common to multiple phases to the trajectory
        traj.add_parameter('m', val=174200., opt=False, units='lbm',
                           desc='aircraft mass',
                           targets={'br_to_v1': ['m'], 'v1_to_vr': ['m'], 'rto': ['m'],
                                    'rotate': ['m'], 'climb': ['m']})

        traj.add_parameter('T_nominal', val=27000 * 2, opt=False, units='lbf', static_targets=True,
                           desc='nominal aircraft thrust',
                           targets={'br_to_v1': ['T']})

        traj.add_parameter('T_engine_out', val=27000, opt=False, units='lbf', static_targets=True,
                           desc='thrust under a single engine',
                           targets={'v1_to_vr': ['T'], 'rotate': ['T'], 'climb': ['T']})

        traj.add_parameter('T_shutdown', val=0.0, opt=False, units='lbf', static_targets=True,
                           desc='thrust when engines are shut down for rejected takeoff',
                           targets={'rto': ['T']})

        traj.add_parameter('mu_r_nominal', val=0.03, opt=False, units=None, static_targets=True,
                           desc='nominal runway friction coeffcient',
                           targets={'br_to_v1': ['mu_r'], 'v1_to_vr': ['mu_r'], 'rotate': ['mu_r']})

        traj.add_parameter('mu_r_braking', val=0.3, opt=False, units=None, static_targets=True,
                           desc='runway friction coefficient under braking',
                           targets={'rto': ['mu_r']})

        traj.add_parameter('h_runway', val=0., opt=False, units='ft', static_targets=False,
                           desc='runway altitude',
                           targets={'br_to_v1': ['h'], 'v1_to_vr': ['h'], 'rto': ['h'],
                                    'rotate': ['h']})

        traj.add_parameter('rho', val=1.225, opt=False, units='kg/m**3', static_targets=True,
                           desc='atmospheric density',
                           targets={'br_to_v1': ['rho'], 'v1_to_vr': ['rho'], 'rto': ['rho'],
                                    'rotate': ['rho']})

        traj.add_parameter('S', val=124.7, opt=False, units='m**2', static_targets=True,
                           desc='aerodynamic reference area',
                           targets={'br_to_v1': ['S'], 'v1_to_vr': ['S'], 'rto': ['S'],
                                    'rotate': ['S'], 'climb': ['S']})

        traj.add_parameter('CD0', val=0.03, opt=False, units=None, static_targets=True,
                           desc='zero-lift drag coefficient',
                           targets={f'{phase}': ['CD0'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('AR', val=9.45, opt=False, units=None, static_targets=True,
                           desc='wing aspect ratio',
                           targets={f'{phase}': ['AR'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                     'rto', 'rotate', 'climb']})

        traj.add_parameter('e', val=801, opt=False, units=None, static_targets=True,
                           desc='Oswald span efficiency factor',
                           targets={f'{phase}': ['e'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                    'rto', 'rotate', 'climb']})

        traj.add_parameter('span', val=35.7, opt=False, units='m', static_targets=True,
                           desc='wingspan',
                           targets={f'{phase}': ['span'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                       'rto', 'rotate', 'climb']})

        traj.add_parameter('h_w', val=1.0, opt=False, units='m', static_targets=True,
                           desc='height of wing above CG',
                           targets={f'{phase}': ['h_w'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('CL0', val=0.5, opt=False, units=None, static_targets=True,
                           desc='zero-alpha lift coefficient',
                           targets={f'{phase}': ['CL0'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('CL_max', val=2.0, opt=False, units=None, static_targets=True,
                           desc='maximum lift coefficient for linear fit',
                           targets={f'{phase}': ['CL_max'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                         'rto', 'rotate', 'climb']})

        # Standard "end of first phase to beginning of second phase" linkages
        traj.link_phases(['br_to_v1', 'v1_to_vr'], vars=['time', 'r', 'v'])
        traj.link_phases(['v1_to_vr', 'rotate'], vars=['time', 'r', 'v', 'alpha'])
        traj.link_phases(['rotate', 'climb'], vars=['time', 'r', 'v', 'alpha'])
        traj.link_phases(['br_to_v1', 'rto'], vars=['time', 'r', 'v'])

        # Less common "final value of r must be the match at ends of two phases".
        traj.add_linkage_constraint(phase_a='rto', var_a='r', loc_a='final',
                                    phase_b='climb', var_b='r', loc_b='final',
                                    ref=1000,
                                    connected=False)

        # Define the constraints and objective for the optimal control problem
        rto.add_boundary_constraint('v', loc='final', upper=0.001, ref=100, linear=True)

        rotate.add_boundary_constraint('F_r', loc='final', equals=0, ref=100000)

        climb.add_boundary_constraint('h', loc='final', equals=35, ref=35, units='ft', linear=True)
        climb.add_boundary_constraint('gam', loc='final', equals=5, ref=5, units='deg', linear=True)
        climb.add_path_constraint('gam', lower=0, upper=5, ref=5, units='deg')
        climb.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.2, ref=1.2)

        rto.add_objective('r', loc='final', ref=1000.0)

        #
        # Setup the problem and set the initial guess
        #
        p.setup(check=True)

        p.set_val('traj.br_to_v1.t_initial', 0)
        p.set_val('traj.br_to_v1.t_duration', 35)
        p.set_val('traj.br_to_v1.states:r', br_to_v1.interp('r', [0, 2500.0]))
        p.set_val('traj.br_to_v1.states:v', br_to_v1.interp('v', [0.0001, 100.0]))
        p.set_val('traj.br_to_v1.parameters:alpha', 0, units='deg')

        p.set_val('traj.v1_to_vr.t_initial', 35)
        p.set_val('traj.v1_to_vr.t_duration', 35)
        p.set_val('traj.v1_to_vr.states:r', v1_to_vr.interp('r', [2500, 300.0]))
        p.set_val('traj.v1_to_vr.states:v', v1_to_vr.interp('v', [100, 110.0]))
        p.set_val('traj.v1_to_vr.parameters:alpha', 0.0, units='deg')

        p.set_val('traj.rto.t_initial', 35)
        p.set_val('traj.rto.t_duration', 1)
        p.set_val('traj.rto.states:r', rto.interp('r', [2500, 5000.0]))
        p.set_val('traj.rto.states:v', rto.interp('v', [110, 0.0001]))
        p.set_val('traj.rto.parameters:alpha', 0.0, units='deg')

        p.set_val('traj.rotate.t_initial', 35)
        p.set_val('traj.rotate.t_duration', 5)
        p.set_val('traj.rotate.states:r', rotate.interp('r', [1750, 1800.0]))
        p.set_val('traj.rotate.states:v', rotate.interp('v', [80, 85.0]))
        p.set_val('traj.rotate.controls:alpha', 0.0, units='deg')

        p.set_val('traj.climb.t_initial', 30)
        p.set_val('traj.climb.t_duration', 20)
        p.set_val('traj.climb.states:r', climb.interp('r', [5000, 5500.0]), units='ft')
        p.set_val('traj.climb.states:v', climb.interp('v', [160, 170.0]), units='kn')
        p.set_val('traj.climb.states:h', climb.interp('h', [0, 35.0]), units='ft')
        p.set_val('traj.climb.states:gam', climb.interp('gam', [0, 5.0]), units='deg')
        p.set_val('traj.climb.controls:alpha', 5.0, units='deg')

        p.run_model()

        model_data = _create_model_data(traj)

        test_dir = pathlib.Path(__file__).parents[0]

        # Uncomment to regenerate baseline data
        # with open(test_dir / 'model_data.dat', 'w') as f:
        #     json.dump(model_data, f)

        with open(test_dir / 'model_data.dat', 'r') as f:
            reference_data = json.load(f)

        flat_model_data = _flatten_dict(model_data)
        flat_ref_data = _flatten_dict(reference_data)

        self.maxDiff = None
        self.assertDictEqual(flat_model_data, flat_ref_data)


if __name__ == "__main__":
    unittest.main()



import openmdao.api as om


class DoubleIntegratorODE(om.ExplicitComponent):
    """
    The double integrator is a special case where the state rates are all set to other states
    or parameters.  Since we aren't computing any other outputs, the ODE doesn't actually
    need to compute anything.  OpenMDAO will warn us that the component has no outputs, but
    Dymos will solve the problem just fine.

    Note we still have to declare the num_nodes option in initialize so that Dymos can instantiate
    the ODE.

    Also note that neither time, states, nor parameters have targets, since there are no inputs
    in the ODE system.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)



from .phase import Phase
from ..transcriptions import ExplicitShooting


class SimulationPhase(Phase):
    """
    A special phase class for performing simulation of an ODE without derivatives.

    Note: For the arguments which accept _unspecified, any argument not provided will be overridden
    by the simulation_options of the phase specified by `from_phase`.

    Parameters
    ----------
    transcription : ExplicitShooting
        The transcription used for the SimulationPhase. It must be an instance of ExplicitShooting.

    **kwargs : dict
        Dictionary of optional phase arguments.
    """
    def __init__(self, transcription=None, **kwargs):
        if not isinstance(transcription, ExplicitShooting):
            raise ValueError('The transcription for a SimulationPhase must be '
                             'ExplicitShooting. Use Phase.get_simulation_phase()'
                             'to create a simulation Phase.')
        super().__init__(transcription=transcription, **kwargs)

    def duplicate(self, *args, **kwargs):
        """
        Create a copy of this phase where most options and attributes are deep copies of those in the original.

        By default, a deepcopy of the transcription in the original phase is used.
        Boundary constraints, path constraints, and objectives are _NOT_ copied by default, but the user may opt to do so.
        By default, initial time is not fixed, nor are the initial or final state values.
        These also can be overridden with the appropriate arguments.

        Parameters
        ----------
        *args
            Additional arguments.
        **kwargs
            Keyword arguments.

        Raises
        ------
        NotImplmentedError
            This method is not yet supported for SimulationPhase
        """
        raise NotImplementedError('SimulationPhase does not support the duplicate method.')

    def set_vals_from_phase(self, from_phase):
        """
        Set the necessary values to simulate the phase based on variables in the given phase.

        Parameters
        ----------
        from_phase : Phase
            The dymos phase from which this simulation phase should pull its values.
        """
        # The use of `from_src=False` in the get_val calls here is due to the fact that the input/output
        # vectors are in `from_phase` are already populated and we don't need to track these values
        # to their ultimate source.

        t_initial = from_phase.get_val('t_initial', units=self.time_options['units'], from_src=False)
        self.set_val('t_initial', t_initial, units=self.time_options['units'])

        t_duration = from_phase.get_val('t_duration', units=self.time_options['units'], from_src=False)
        self.set_val('t_duration', t_duration, units=self.time_options['units'])

        avail_io = {meta['prom_name'] for meta in
                    from_phase.get_io_metadata(iotypes=('input', 'output'), get_remote=True).values()}

        for name, options in self.state_options.items():
            if f'states:{name}' in avail_io:
                val = from_phase.get_val(f'states:{name}', units=options['units'], from_src=False)[0, ...]
            elif f'initial_states:{name}' in avail_io:
                val = from_phase.get_val(f'initial_states:{name}', units=options['units'], from_src=False)
            else:
                raise RuntimeError('Unable to find state values in original phase')
            self.set_val(f'initial_states:{name}', val, units=options['units'])

        for name, options in self.parameter_options.items():
            val = from_phase.get_val(f'parameters:{name}', units=options['units'], from_src=False)
            self.set_val(f'parameters:{name}', val, units=options['units'])

        for name, options in self.control_options.items():
            val = from_phase.get_val(f'controls:{name}', units=options['units'], from_src=False)
            self.set_val(f'controls:{name}', val, units=options['units'])

    def add_boundary_constraint(self, name, loc, constraint_name=None, units=None,
                                shape=None, indices=None, lower=None, upper=None, equals=None,
                                scaler=None, adder=None, ref=None, ref0=None, linear=False, flat_indices=False):
        r"""
        Add a boundary constraint to a variable in the phase.

        Parameters
        ----------
        name : str
            Name of the variable to constrain. May also provide an expression to be evaluated and constrained.
            If a single variable and the name is not a state, control, or 'time',
            then this is assumed to be the path of the variable to be constrained in the ODE.
            If an expression, it must be provided in the form of an equation with a left- and right-hand side.
        loc : str
            The location of the boundary constraint ('initial' or 'final').
        constraint_name : str or None
            The name of the boundary constraint. By default, this is 'var_constraint' if name is a single variable,
             or the left-hand side of the equation if name is an expression.
        units : str or None
            The units in which the boundary constraint is to be applied.  If None, use the
            units associated with the constrained output.  If provided, must be compatible with
            the variables units.
        shape : tuple, list, ndarray, or None
            The shape of the variable being boundary-constrained.  This can be inferred
            automatically for time, states, controls, and parameters, but is required
            if the constrained variable is an output of the ODE system.
        indices : tuple, list, ndarray, slice, or None
            The indices of the output variable to be boundary constrained at either the initial or final time in the
            phase. When the variable is multi-dimensional, this should be a list of lists, one for each dimension,
            containing the indices to be constrained.  Note the behavior of indices changes depending on the value
            of the flat_indices option.
        lower : float or ndarray, optional
            Lower boundary for the variable.
        upper : float or ndarray, optional
            Upper boundary for the variable.
        equals : float or ndarray, optional
            Equality constraint value for the variable.
        scaler : float or ndarray, optional
            Value to multiply the model value to get the scaled value. Scaler
            is second in precedence.
        adder : float or ndarray, optional
            Value to add to the model value to get the scaled value. Adder
            is first in precedence.
        ref : float or ndarray, optional
            Value of response variable that scales to 1.0 in the driver.
        ref0 : float or ndarray, optional
            Value of response variable that scales to 0.0 in the driver.
        linear : bool
            Set to True if constraint is linear. Setting this to True when the constraint is not a linear function
            of the design variables will result in a failure of the optimization.
        flat_indices : bool
            If True, treat indices as flattened C-ordered indices of elements to constrain. Otherwise,
            indices should be a tuple or list giving the elements to constrain at each point in time.
        """
        raise NotImplementedError('SimulationPhase does not support boundary constraints.')

    def add_path_constraint(self, name, constraint_name=None, units=None, shape=None, indices=None,
                            lower=None, upper=None, equals=None, scaler=None, adder=None, ref=None,
                            ref0=None, linear=False, flat_indices=False):
        r"""
        Add a path constraint to a variable in the phase.

        Parameters
        ----------
        name : str
            Name of the variable to constrain. May also provide an expression to be evaluated and constrained.
            If a single variable and the name is not a state, control, or 'time',
            then this is assumed to be the path of the variable to be constrained in the ODE.
            If an expression, it must be provided in the form of an equation with a left- and right-hand side.
        constraint_name : str or None
            The name of the path constraint. By default, this is 'var_constraint' if name is a single variable,
             or the left-hand side of the equation if name is an expression.
        units : str or None
            The units in which the boundary constraint is to be applied.  If None, use the
            units associated with the constrained output.  If provided, must be compatible with
            the variables units.
        shape : tuple, list, ndarray, or None
            The shape of the variable being boundary-constrained.  This can be inferred
            automatically for time, states, controls, and parameters, but is required
            if the constrained variable is an output of the ODE system.
        indices : tuple, list, ndarray, or None
            The indices of the output variable to be constrained at each point in time in the phase.
            When the variable is multi-dimensional, this should be a list of lists, one for each dimension,
            containing the indices to be constrained.  Note the behavior of indices changes depending on the value
            of the flat_indices option.
        lower : float or ndarray, optional
            Lower boundary for the variable.
        upper : float or ndarray, optional
            Upper boundary for the variable.
        equals : float or ndarray, optional
            Equality constraint value for the variable.
        scaler : float or ndarray, optional
            Value to multiply the model value to get the scaled value. Scaler
            is second in precedence.
        adder : float or ndarray, optional
            Value to add to the model value to get the scaled value. Adder
            is first in precedence.
        ref : float or ndarray, optional
            Value of response variable that scales to 1.0 in the driver.
        ref0 : float or ndarray, optional
            Value of response variable that scales to 0.0 in the driver.
        linear : bool
            Set to True if constraint is linear. If set to True and the constrained output is not a linear function
            of the design variables, the optimization will fail.
        flat_indices : bool
            If True, treat indices as flattened C-ordered indices of elements to constrain at each given point in time.
            Otherwise, indices should be a tuple or list giving the elements to constrain at each point in time.
        """
        raise NotImplementedError('SimulationPhase does not support path constraints.')

    def add_objective(self, name, loc='final', index=None, shape=(1,), units=None, ref=None, ref0=None,
                      adder=None, scaler=None, parallel_deriv_color=None):
        """
        Add an objective in the phase.

        If name is not a state, control, control rate, or 'time', then this is assumed to be the
        path of the variable to be constrained in the RHS.

        Parameters
        ----------
        name : str
            Name of the objective variable.  This should be one of the integration variable, a state or control
            variable, the path to an output from the top level of the RHS, or an expression to be evaluated.
            If an expression, it must be provided in the form of an equation with a left- and right-hand side.
        loc : str
            Where in the phase the objective is to be evaluated.  Valid
            options are 'initial' and 'final'.  The default is 'final'.
        index : int, optional
            If variable is an array at each point in time, this indicates which index is to be
            used as the objective, assuming C-ordered flattening.
        shape : int, optional
            The shape of the objective variable, at a point in time.
        units : str, optional
            The units of the objective function.  If None, use the units associated with the target.
            If provided, must be compatible with the target units.
        ref : float or ndarray, optional
            Value of response variable that scales to 1.0 in the driver.
        ref0 : float or ndarray, optional
            Value of response variable that scales to 0.0 in the driver.
        adder : float or ndarray, optional
            Value to add to the model value to get the scaled value. Adder
            is first in precedence.
        scaler : float or ndarray, optional
            Value to multiply the model value to get the scaled value. Scaler
            is second in precedence.
        parallel_deriv_color : str
            If specified, this design var will be grouped for parallel derivative
            calculations with other variables sharing the same parallel_deriv_color.
        """
        raise NotImplementedError('SimulationPhase does not support optimization objectives.')

    def check_time_options(self):
        """
        Check that time options are valid and issue warnings if invalid options are provided.

        This check is not performed by SimulationPhase.

        Warns
        -----
        RuntimeWarning
            RuntimeWarning is issued in the case of one or more invalid time options.
        """
        pass

    def _check_control_options(self):
        """
        Check that control options are valid and issue warnings if invalid options are provided.

        This check is not performed by SimulationPhase.

        Warns
        -----
        RuntimeWarning
            RuntimeWarning is issued in the case of one or more invalid time options.
        """
        pass

    def _check_parameter_options(self):
        """
        Check that parameter options are valid and issue warnings if invalid options are provided.

        This check is not performed by SimulationPhase.

        Warns
        -----
        RuntimeWarning
            RuntimeWarning is issued in the case of one or more invalid time options.
        """
        pass



import importlib
import os
import pathlib
import unittest
from numpy.testing import assert_almost_equal

import sys

from openmdao.utils.general_utils import set_pyoptsparse_opt, printoptions
from openmdao.utils.testing_utils import use_tempdirs, set_env_vars_context
from openmdao.utils.tests.test_hooks import hooks_active

import dymos as dm
import dymos.examples.brachistochrone.test.ex_brachistochrone_vector_states as ex_brachistochrone_vs
from dymos.utils.testing_utils import assert_check_partials, _get_reports_dir

bokeh_available = importlib.util.find_spec('bokeh') is not None

OPT, OPTIMIZER = set_pyoptsparse_opt('SNOPT')


@use_tempdirs
class TestBrachistochroneVectorStatesExample(unittest.TestCase):

    def setUp(self):
        self.testflo_running = os.environ.pop('TESTFLO_RUNNING', None)

    def tearDown(self):
        # restore what was there before running the test
        if self.testflo_running is not None:
            os.environ['TESTFLO_RUNNING'] = self.testflo_running

    def assert_results(self, p):
        t_initial = p.get_val('traj0.phase0.timeseries.time')[0]
        t_final = p.get_val('traj0.phase0.timeseries.time')[-1]

        x0 = p.get_val('traj0.phase0.timeseries.pos')[0, 0]
        xf = p.get_val('traj0.phase0.timeseries.pos')[0, -1]

        y0 = p.get_val('traj0.phase0.timeseries.pos')[0, 1]
        yf = p.get_val('traj0.phase0.timeseries.pos')[-1, 1]

        v0 = p.get_val('traj0.phase0.timeseries.v')[0, 0]
        vf = p.get_val('traj0.phase0.timeseries.v')[-1, 0]

        g = p.get_val('traj0.phase0.parameter_vals:g')

        thetaf = p.get_val('traj0.phase0.timeseries.theta')[-1, 0]

        assert_almost_equal(t_initial, 0.0)
        assert_almost_equal(x0, 0.0)
        assert_almost_equal(y0, 10.0)
        assert_almost_equal(v0, 0.0)

        assert_almost_equal(t_final, 1.8016, decimal=4)
        assert_almost_equal(xf, 10.0, decimal=3)
        assert_almost_equal(yf, 5.0, decimal=3)
        assert_almost_equal(vf, 9.902, decimal=3)
        assert_almost_equal(g, 9.80665, decimal=3)

        assert_almost_equal(thetaf, 100.12, decimal=0)

    def assert_partials(self, p):
        with printoptions(linewidth=1024, edgeitems=100):
            cpd = p.check_partials(method='cs', compact_print=True, out_stream=None)
        assert_check_partials(cpd)

    def test_ex_brachistochrone_vs_radau_compressed(self):
        ex_brachistochrone_vs.SHOW_PLOTS = True
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='radau-ps',
                                                           compressed=True,
                                                           force_alloc_complex=True,
                                                           run_driver=False)
        p.run_driver()
        self.assert_partials(p)

    def test_ex_brachistochrone_vs_radau_uncompressed(self):
        ex_brachistochrone_vs.SHOW_PLOTS = True
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='radau-ps',
                                                           compressed=False,
                                                           force_alloc_complex=True,
                                                           run_driver=True)
        self.assert_results(p)
        self.assert_partials(p)

    def test_ex_brachistochrone_vs_gl_compressed(self):
        ex_brachistochrone_vs.SHOW_PLOTS = True
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='gauss-lobatto',
                                                           compressed=True,
                                                           force_alloc_complex=True,
                                                           run_driver=True)

        self.assert_results(p)
        self.assert_partials(p)

    def test_ex_brachistochrone_vs_gl_uncompressed(self):
        ex_brachistochrone_vs.SHOW_PLOTS = True
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='gauss-lobatto',
                                                           transcription_order=5,
                                                           compressed=False,
                                                           force_alloc_complex=True,
                                                           run_driver=True)
        self.assert_results(p)
        self.assert_partials(p)

    def test_ex_brachistochrone_vs_birkhoff(self):
        ex_brachistochrone_vs.SHOW_PLOTS = True
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='birkhoff',
                                                           transcription_order=12,
                                                           compressed=True,
                                                           force_alloc_complex=True,
                                                           run_driver=True)

        self.assert_results(p)
        self.assert_partials(p)

    @unittest.skipIf(not bokeh_available, 'bokeh unavailable')
    @hooks_active
    def test_bokeh_plots(self):
        with set_env_vars_context(OPENMDAO_REPORTS='1'):
            with dm.options.temporary(plots='bokeh'):
                p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='radau-ps',
                                                                   compressed=False,
                                                                   force_alloc_complex=True,
                                                                   run_driver=True,
                                                                   simulate=True,
                                                                   make_plots=True)

                self.assert_results(p)
                self.assert_partials(p)

                html_file = pathlib.Path(_get_reports_dir(p)) / 'traj0_results_report.html'
                self.assertTrue(html_file.exists(), msg=f'{html_file} does not exist!')

                with open(html_file) as f:
                    html_data = f.read()

                expected_labels = ['"axis_label":"pos[0] (m)"',
                                   '"axis_label":"pos[1] (m)"',
                                   '"axis_label":"v (m/s)"',
                                   '"axis_label":"theta (deg)"']

                for label in expected_labels:
                    self.assertIn(label, html_data)

                self.assertNotIn('"axis_label":"pos (m)"', html_data)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import openmdao.api as om
import dymos as dm

from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal


class ODEComp(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('h', shape=nn, units='m')
        self.add_input('v', shape=nn, units='m/s')
        self.add_output('hdot', shape=nn, units='m/s')
        self.add_output('vdot', shape=nn, units='m/s**2')

        self.declare_partials(of='hdot', wrt='v', rows=np.arange(nn), cols=np.arange(nn), val=1.0)

    def compute(self, inputs, outputs):
        outputs['hdot'] = inputs['v']
        outputs['vdot'] = -9.80665


@use_tempdirs
class TestPhaseLinkageComp(unittest.TestCase):

    @staticmethod
    def make_problem(link_all_vars=False, connected=True):
        dm.options['include_check_partials'] = True
        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        traj = dm.Trajectory()
        p.model.add_subsystem('traj', traj)

        t = dm.Radau(num_segments=1, order=3, compressed=True)
        phase0 = dm.Phase(ode_class=ODEComp, transcription=t)

        traj.add_phase('phase0', phase0)
        phase0.set_time_options(fix_initial=True, fix_duration=True)

        phase0.add_state('h', fix_initial=True, fix_final=False,
                         solve_segments='forward', rate_source='hdot')
        phase0.add_state('v', fix_initial=True, fix_final=False,
                         solve_segments='forward', rate_source='vdot')

        phase1 = dm.Phase(ode_class=ODEComp, transcription=t)

        traj.add_phase('phase1', phase1)
        phase1.set_time_options(input_initial=connected, fix_duration=True)

        phase1.add_state('h', input_initial=connected, fix_final=False,
                         solve_segments='forward', rate_source='hdot')
        phase1.add_state('v', input_initial=connected, fix_final=False,
                         solve_segments='forward', rate_source='vdot')

        linked_vars = ['*'] if link_all_vars else ['time', 'h', 'v']
        traj.link_phases(['phase0', 'phase1'], vars=linked_vars, connected=connected)

        phase1.add_objective('h', loc='final')

        p.setup()

        p.set_val('traj.phase0.t_initial', 0.0)
        p.set_val('traj.phase0.t_duration', 1.0)
        p.set_val('traj.phase0.states:h', phase0.interp('h', [19.6133, 15]))
        p.set_val('traj.phase0.states:v', phase0.interp('v', [0, -5]))

        p.set_val('traj.phase1.t_duration', 1.0)

        return p

    def test_link_phases_specified_vars_unconnected(self):
        p = self.make_problem(connected=False)
        dm.run_problem(p, run_driver=True)

        h0 = p.get_val('traj.phase0.states:h')[-1]
        h1 = p.get_val('traj.phase1.states:h')[0]
        v0 = p.get_val('traj.phase0.states:v')[-1]
        v1 = p.get_val('traj.phase1.states:v')[0]

        assert_near_equal(h0, h1)
        assert_near_equal(v0, v1)

    def test_link_phases_all_vars_unconnected(self):
        p = self.make_problem(link_all_vars=True, connected=False)
        dm.run_problem(p, run_driver=True)

        h0 = p.get_val('traj.phase0.states:h')[-1]
        h1 = p.get_val('traj.phase1.states:h')[0]
        v0 = p.get_val('traj.phase0.states:v')[-1]
        v1 = p.get_val('traj.phase1.states:v')[0]

        assert_near_equal(h0, h1)
        assert_near_equal(v0, v1)

    def test_link_phases_specified_vars_connected(self):
        p = self.make_problem(connected=True)
        dm.run_problem(p, run_driver=False)

        h0 = p.get_val('traj.phase0.states:h')[-1]
        h1 = p.get_val('traj.phase1.states:h')[0]
        v0 = p.get_val('traj.phase0.states:v')[-1]
        v1 = p.get_val('traj.phase1.states:v')[0]

        assert_near_equal(h0, h1)
        assert_near_equal(v0, v1)

    def test_link_phases_all_vars_connected(self):
        p = self.make_problem(link_all_vars=True, connected=True)
        dm.run_problem(p, run_driver=False)

        h0 = p.get_val('traj.phase0.states:h')[-1]
        h1 = p.get_val('traj.phase1.states:h')[0]
        v0 = p.get_val('traj.phase0.states:v')[-1]
        v1 = p.get_val('traj.phase1.states:v')[0]

        assert_near_equal(h0, h1)
        assert_near_equal(v0, v1)



import openmdao.api as om
import numpy as np


class RocketODE(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('h', val=np.zeros(nn), units='ft')
        self.add_input('v', val=np.zeros(nn), units='ft/s')
        self.add_input('m', val=np.zeros(nn), units='slug')
        self.add_input('T', val=np.zeros(nn), units='lbf')

        self.add_output('h_dot', val=np.zeros(nn), units='ft/s')
        self.add_output('v_dot', val=np.zeros(nn), units='ft/s**2')
        self.add_output('m_dot', val=np.zeros(nn), units='slug/s')

        ar = np.arange(nn)
        self.declare_partials(of='h_dot', wrt='v', rows=ar, cols=ar, val=1.0)
        self.declare_partials(of='v_dot', wrt='h', rows=ar, cols=ar)
        self.declare_partials(of='v_dot', wrt='v', rows=ar, cols=ar)
        self.declare_partials(of='v_dot', wrt='m', rows=ar, cols=ar)
        self.declare_partials(of='v_dot', wrt='T', rows=ar, cols=ar)

        self.declare_partials(of='m_dot', wrt='T', rows=ar, cols=ar, val=-1/1580.9425)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        h = inputs['h']
        v = inputs['v']
        m = inputs['m']
        T = inputs['T']

        sigma = 5.4915e-5
        g = 32.174
        c = 1580.9425
        h0 = 23800

        outputs['h_dot'] = v
        outputs['v_dot'] = (T - sigma * v**2 * np.exp(-h/h0))/m - g
        outputs['m_dot'] = -T / c

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        h = inputs['h']
        v = inputs['v']
        m = inputs['m']
        T = inputs['T']

        sigma = 5.4915e-5
        h0 = 23800

        partials['v_dot', 'h'] = (sigma * v**2 * np.exp(-h/h0))/(m * h0)
        partials['v_dot', 'v'] = -(2 * sigma * v * np.exp(-h/h0)) / m
        partials['v_dot', 'm'] = -(T - sigma * v**2 * np.exp(-h/h0)) / (m**2)
        partials['v_dot', 'T'] = 1 / m



import unittest
from numpy.testing import assert_almost_equal

import openmdao.api as om

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.utils.testing_utils import use_tempdirs
OPT, OPTIMIZER = set_pyoptsparse_opt('SNOPT', fallback=True)


@use_tempdirs
class TestBrachistochroneRefineGrid(unittest.TestCase):

    def make_problem(self, transcription='radau-ps', num_segments=5, transcription_order=3, compressed=True):

        self.p = p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        if transcription == 'gauss-lobatto':
            t = dm.GaussLobatto(num_segments=num_segments,
                                order=transcription_order,
                                compressed=compressed)
        elif transcription == 'radau-ps':
            t = dm.Radau(num_segments=num_segments,
                         order=transcription_order,
                         compressed=compressed)
        elif transcription == 'birkhoff':
            t = dm.Birkhoff(num_nodes=transcription_order+1,
                            grid_type='cgl')

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)

        phase.set_refine_options(refine=True, tol=1.0E-6)

        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj0', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])

        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        return p

    def run_asserts(self, p):

        t_initial = p.get_val('traj0.phase0.timeseries.time')[0]
        tf = p.get_val('traj0.phase0.timeseries.time')[-1]

        x0 = p.get_val('traj0.phase0.timeseries.x')[0]
        xf = p.get_val('traj0.phase0.timeseries.x')[-1]

        y0 = p.get_val('traj0.phase0.timeseries.y')[0]
        yf = p.get_val('traj0.phase0.timeseries.y')[-1]

        v0 = p.get_val('traj0.phase0.timeseries.v')[0]
        vf = p.get_val('traj0.phase0.timeseries.v')[-1]

        g = p.get_val('traj0.phase0.parameter_vals:g')[0]

        thetaf = p.get_val('traj0.phase0.timeseries.theta')[-1]

        assert_almost_equal(t_initial, 0.0)
        assert_almost_equal(x0, 0.0)
        assert_almost_equal(y0, 10.0)
        assert_almost_equal(v0, 0.0)

        assert_almost_equal(tf, 1.8016, decimal=4)
        assert_almost_equal(xf, 10.0, decimal=3)
        assert_almost_equal(yf, 5.0, decimal=3)
        assert_almost_equal(vf, 9.902, decimal=3)
        assert_almost_equal(g, 9.80665, decimal=3)

        assert_almost_equal(thetaf, 100.12, decimal=0)

    def test_refine_brachistochrone_radau_compressed(self):
        p = self.make_problem(transcription='radau-ps', num_segments=5, transcription_order=3, compressed=True)
        dm.run_problem(p, refine_iteration_limit=3)
        self.run_asserts(self.p)

    def test_refine_brachistochrone_gauss_lobatto_compressed(self):
        p = self.make_problem(transcription='gauss-lobatto', num_segments=5, transcription_order=3, compressed=True)
        dm.run_problem(p, refine_iteration_limit=3)
        self.run_asserts(self.p)

    def test_refine_brachistochrone_birkhoff_compressed(self):
        p = self.make_problem(transcription='birkhoff', num_segments=1, transcription_order=12, compressed=True)
        dm.run_problem(p)
        self.run_asserts(self.p)



"""Unit Tests for the code that does automatic report generation"""
import unittest
import os

import openmdao
import openmdao.api as om
import openmdao.core.problem
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.tests.test_hooks import hooks_active
from openmdao.visualization.n2_viewer.n2_viewer import _default_n2_filename
from openmdao.visualization.scaling_viewer.scaling_report import _default_scaling_filename


import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from dymos.utils.misc import om_version


def setup_model_radau(do_reports, probname):
    p = om.Problem(model=om.Group(), name=probname)

    p.driver = om.ScipyOptimizeDriver()
    p.driver.declare_coloring(tol=1.0E-12)

    t = dm.Radau(num_segments=10, order=3)

    traj = dm.Trajectory()
    phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)
    p.model.add_subsystem('traj0', traj)
    traj.add_phase('phase0', phase)

    # p.model.add_subsystem('traj0', traj)

    phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

    phase.add_state('x', fix_initial=True, fix_final=False)
    phase.add_state('y', fix_initial=True, fix_final=False)

    # Note that by omitting the targets here Dymos will automatically attempt to connect
    # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
    phase.add_state('v', fix_initial=True, fix_final=False)

    phase.add_control('theta', continuity=True, rate_continuity=True,
                      units='deg', lower=0.01, upper=179.9)

    phase.add_parameter('g', targets=['g'], units='m/s**2')

    phase.add_boundary_constraint('x', loc='final', equals=10)
    phase.add_boundary_constraint('y', loc='final', equals=5)
    # Minimize time at the end of the phase
    phase.add_objective('time_phase', loc='final', scaler=10)

    p.setup()

    phase.set_simulate_options(method='RK23')

    phase.set_time_val(initial=0.0, duration=2.0)

    phase.set_state_val('x', [0, 10])
    phase.set_state_val('y', [10, 5])
    phase.set_state_val('v', [0, 9.9])

    phase.set_control_val('theta', [5, 100.5])
    phase.set_parameter_val('g', 9.80665)

    dm.run_problem(p, run_driver=True, simulate=True, simulate_kwargs={'reports': do_reports})

    return p


def setup_model_shooting(do_reports, probname):
    prob = om.Problem(name=probname)

    prob.driver = om.ScipyOptimizeDriver()
    prob.driver.declare_coloring(tol=1.0E-12)

    tx = dm.ExplicitShooting(grid=dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=6, compressed=False),
                             subprob_reports=do_reports)

    phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

    phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

    # automatically discover states
    phase.set_state_options('x', fix_initial=True)
    phase.set_state_options('y', fix_initial=True)
    phase.set_state_options('v', fix_initial=True)

    phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1, upper=9.80665)
    phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9,
                      ref=90., rate2_continuity=True)

    phase.add_boundary_constraint('x', loc='final', equals=10.0)
    phase.add_boundary_constraint('y', loc='final', equals=5.0)

    prob.model.add_subsystem('phase0', phase)

    phase.add_objective('time', loc='final')

    prob.setup(force_alloc_complex=True)

    phase.set_time_val(initial=0.0, duration=2.0)

    phase.set_state_val('x', [0, 10])
    phase.set_state_val('y', [10, 5])
    phase.set_state_val('v', [0, 9.9])

    phase.set_control_val('theta', [0.01, 90], units='deg')
    phase.set_parameter_val('g', 1.0)

    dm.run_problem(prob, run_driver=True, simulate=True)

    return prob


@use_tempdirs
class TestSubproblemReportToggle(unittest.TestCase):

    def setUp(self):
        self.n2_filename = _default_n2_filename
        self.scaling_filename = _default_scaling_filename

        # set things to a known initial state for all the test runs
        openmdao.core.problem._problem_names = []  # need to reset these to simulate separate runs
        os.environ.pop('OPENMDAO_REPORTS', None)
        os.environ.pop('OPENMDAO_REPORTS_DIR', None)
        # We need to remove the TESTFLO_RUNNING environment variable for these tests to run.
        # The reports code checks to see if TESTFLO_RUNNING is set and will not do anything if set
        # But we need to remember whether it was set so we can restore it
        self.testflo_running = os.environ.pop('TESTFLO_RUNNING', None)

        self.count = 0

    def tearDown(self):
        # restore what was there before running the test
        if self.testflo_running is not None:
            os.environ['TESTFLO_RUNNING'] = self.testflo_running

    @unittest.skipIf(om_version()[0] <= (3, 34, 2), 'Requires OpenMDAO version later than 3.34.2')
    @hooks_active
    def test_no_sim_reports(self):
        p = setup_model_radau(do_reports=False, probname='test_no_sim_reports')

        main_outputs_dir = p.get_outputs_dir()

        sim_outputs_dir = main_outputs_dir / 'traj0_simulation_out'
        sim_reports_dir = sim_outputs_dir / 'reports'

        self.assertFalse(sim_reports_dir.exists())

    @unittest.skipIf(om_version()[0] <= (3, 34, 2), 'Requires OpenMDAO version later than 3.34.2')
    @hooks_active
    def test_make_sim_reports(self):
        p = setup_model_radau(do_reports=True, probname='test_make_sim_reports')

        main_reports_dir = p.get_reports_dir()

        traj = p.model._get_subsystem('traj0')
        sim_reports_dir = traj.sim_prob.get_reports_dir()

        self.assertTrue((main_reports_dir / self.n2_filename).exists())
        self.assertTrue(sim_reports_dir.exists())
        self.assertTrue((sim_reports_dir / self.n2_filename).exists())

    @unittest.skipIf(om_version()[0] <= (3, 34, 2), 'Requires OpenMDAO version later than 3.34.2')
    @hooks_active
    def test_explicitshooting_no_subprob_reports(self):
        p = setup_model_shooting(do_reports=False,
                                 probname='test_explicitshooting_no_subprob_reports')

        main_reports_dir = p.get_reports_dir()
        subprob_reports_dir = p.model.phase0.integrator._eval_subprob.get_reports_dir()

        main_reports = os.listdir(main_reports_dir)

        self.assertFalse(subprob_reports_dir.exists())

        self.assertIn(self.n2_filename, main_reports)
        self.assertIn(self.scaling_filename, main_reports)

    @unittest.skipIf(om_version()[0] <= (3, 34, 2), 'Requires OpenMDAO version later than 3.34.2')
    @hooks_active
    def test_explicitshooting_make_subprob_reports(self):
        p = setup_model_shooting(do_reports=True,
                                 probname='test_explicitshooting_make_subprob_reports')

        main_reports_dir = p.get_reports_dir()
        subprob_reports_dir = p.model.phase0.integrator._eval_subprob.get_reports_dir()

        main_reports = os.listdir(main_reports_dir)
        subprob_reports = os.listdir(subprob_reports_dir)

        self.assertIn(self.n2_filename, main_reports)
        self.assertIn(self.n2_filename, subprob_reports)

        self.assertIn(self.scaling_filename, main_reports)

        # The subprob has no optimization, so should not have a scaling report
        self.assertNotIn(self.scaling_filename, subprob_reports)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import openmdao.api as om
import time
from openmdao.utils.array_utils import evenly_distrib_idxs
from openmdao.utils.mpi import MPI


class VanderpolODE(om.ExplicitComponent):
    """intentionally slow version of vanderpol_ode for effects of demonstrating distributed component calculations

    MPI can run this component in multiple processes, distributing the calculation of derivatives.
    This code has a delay in it to simulate a longer computation. It should run faster with more processes.
    """

    def __init__(self, *args, **kwargs):
        self.progress_prints = False
        super().__init__(*args, **kwargs)

    def initialize(self):
        self.options.declare('num_nodes', types=int)
        self.options.declare('distrib', types=bool, default=False)
        self.options.declare('delay', types=(float,), default=0.0)

    def setup(self):
        nn = self.options['num_nodes']
        comm = self.comm
        rank = comm.rank

        sizes, offsets = evenly_distrib_idxs(comm.size, nn)  # (#cpus, #inputs) -> (size array, offset array)
        self.start_idx = offsets[rank]
        self.io_size = sizes[rank]  # number of inputs and outputs managed by this distributed process
        self.end_idx = self.start_idx + self.io_size

        # inputs: 2 states and a control
        self.add_input('x0', val=np.ones(nn), desc='derivative of Output', units='V/s')
        self.add_input('x1', val=np.ones(nn), desc='Output', units='V')
        self.add_input('u', val=np.ones(nn), desc='control', units=None)

        # outputs: derivative of states
        # the objective function will be treated as a state for computation, so its derivative is an output
        self.add_output('x0dot', val=np.ones(self.io_size), desc='second derivative of Output',
                        units='V/s**2', distributed=self.options['distrib'])
        self.add_output('x1dot', val=np.ones(self.io_size), desc='derivative of Output',
                        units='V/s', distributed=self.options['distrib'])
        self.add_output('Jdot', val=np.ones(self.io_size), desc='derivative of objective',
                        units='1.0/s', distributed=self.options['distrib'])

        # self.declare_coloring(method='cs')
        # # partials
        r = np.arange(self.io_size, dtype=int)
        c = r + self.start_idx

        self.declare_partials(of='x0dot', wrt='x0',  rows=r, cols=c)
        self.declare_partials(of='x0dot', wrt='x1',  rows=r, cols=c)
        self.declare_partials(of='x0dot', wrt='u',   rows=r, cols=c, val=1.0)

        self.declare_partials(of='x1dot', wrt='x0',  rows=r, cols=c, val=1.0)

        self.declare_partials(of='Jdot', wrt='x0',  rows=r, cols=c)
        self.declare_partials(of='Jdot', wrt='x1',  rows=r, cols=c)
        self.declare_partials(of='Jdot', wrt='u',   rows=r, cols=c)

    def compute(self, inputs, outputs):
        # introduce slowness proportional to size of computation
        time.sleep(self.options['delay'] * self.io_size)

        # The inputs contain the entire vector, be each rank will only operate on a portion of it.
        x0 = inputs['x0'][self.start_idx:self.end_idx]
        x1 = inputs['x1'][self.start_idx:self.end_idx]
        u = inputs['u'][self.start_idx:self.end_idx]

        outputs['x0dot'] = (1.0 - x1**2) * x0 - x1 + u
        outputs['x1dot'] = x0
        outputs['Jdot'] = x0**2 + x1**2 + u**2

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):

        # it's necessary to make this component matrix free because the inputs are non-distributed
        # and the outputs are distributed, and the framework doesn't know how to populate the full
        # nondistributed inputs on each rank in reverse mode.

        # FIXME: this delay will be applied on every call to compute_jacvec_product, which may be
        #        more often than was originally intended before this component was converted to
        #        matrix free (it was originally done in compute_partials).
        time.sleep(self.options['delay'])

        myslice = slice(self.start_idx, self.end_idx)

        locx0 = inputs['x0'][myslice]
        locx1 = inputs['x1'][myslice]
        locu = inputs['u'][myslice]

        locdx0 = d_inputs['x0'][myslice]
        locdx1 = d_inputs['x1'][myslice]
        locdu = d_inputs['u'][myslice]

        if mode == 'fwd':
            if 'x0dot' in d_outputs:
                if 'x0' in d_inputs:
                    d_outputs['x0dot'] += (1.0 - locx1**2) * locdx0
                if 'x1' in d_inputs:
                    d_outputs['x0dot'] += (-2.0 * locx1 * locx0 - 1.) * locdx1
                if 'u' in d_inputs:
                    d_outputs['x0dot'] += locdu
            if 'x1dot' in d_outputs:
                if 'x0' in d_inputs:
                    d_outputs['x1dot'] += locdx0
            if 'Jdot' in d_outputs:
                if 'x0' in d_inputs:
                    d_outputs['Jdot'] += 2.0 * locx0 * locdx0
                if 'x1' in d_inputs:
                    d_outputs['Jdot'] += 2.0 * locx1 * locdx1
                if 'u' in d_inputs:
                    d_outputs['Jdot'] += 2.0 * locu * locdu

        elif mode == 'rev':
            if 'x0dot' in d_outputs:
                if 'x0' in d_inputs:
                    d_inputs['x0'][myslice] += (1.0 - locx1**2) * d_outputs['x0dot']
                if 'x1' in d_inputs:
                    d_inputs['x1'][myslice] += (-2.0 * locx1 * locx0 - 1.) * d_outputs['x0dot']
                if 'u' in d_inputs:
                    d_inputs['u'][myslice] += d_outputs['x0dot']
            if 'x1dot' in d_outputs:
                if 'x0' in d_inputs:
                    d_inputs['x0'][myslice] += d_outputs['x1dot']
            if 'Jdot' in d_outputs:
                if 'x0' in d_inputs:
                    d_inputs['x0'][myslice] += 2.0 * locx0 * d_outputs['Jdot']
                if 'x1' in d_inputs:
                    d_inputs['x1'][myslice] += 2.0 * locx1 * d_outputs['Jdot']
                if 'u' in d_inputs:
                    d_inputs['u'][myslice] += 2.0 * locu * d_outputs['Jdot']

            if self.comm.size > 1 and self.options['distrib']:
                d_inputs['x0'] = self.comm.allreduce(d_inputs['x0'], op=MPI.SUM)
                d_inputs['x1'] = self.comm.allreduce(d_inputs['x1'], op=MPI.SUM)
                d_inputs['u'] = self.comm.allreduce(d_inputs['u'], op=MPI.SUM)
